#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.introspection.inspect import entity_properties, is_descriptor
from unittest import TestCase, main, skip
from typing import List


methods_to_run: List[str] = [
    'python_version', 
    'cpuinfo_version', 
    'cpuinfo_version_string', 
    'python_hz_advertised_friendlyversion', 
    'hz_actual_friendly', 
    'hz_advertised', 
    'hz_actual', 
    'arch', 
    'count', 
    'l1_data_cache_size', 
    'l1_instruction_cache_size', 
    'l2_cache_size', 
    'l2_cache_line_size', 
    'l2_cache_associativity', 
    'l3_cache_size', 
    'stepping', 
    'model', 
    'family', 
    'processor_type', 
    'vendor_id_raw', 
    'hardware_raw', 
    'brand_raw', 
    'arch_string_raw', 
    'cores_num', 
    'virtual_cores_num', 
    'l2_cache_size_per_core', 
    'l2_cache_size_per_virtual_core', 
    'l3_cache_size_per_core', 
    'l3_cache_size_per_virtual_core', 
    'is_x86', 
    'is_arm', 
    'flags', 
]


class TestCpuInfo(TestCase):
    def test__run_methods(self):
        from cengal.hardware.info.cpu import cpu_info
        ci = cpu_info()
        for method in methods_to_run:
            print(f'{method}:', getattr(ci, method))
        
        ci.gather_all()
    
    def test__run_methods_2(self):
        from cengal.hardware.info.cpu import cpu_info, CpuInfo
        ci = cpu_info()
        for property in entity_properties(CpuInfo):
            if property.startswith('_'):
                continue

            if is_descriptor(getattr(CpuInfo, property)):
                print(f'{property}:', getattr(ci, property))
        
        ci.gather_all()


if '__main__' == __name__:
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'bigint_to_bytes',
    'bint_to_bytes',
    'bytes_to_bigint',
    'bytes_to_bint',
    'ubigint_to_bytes',
    'ubint_to_bytes',
    'bytes_to_ubigint',
    'bytes_to_ubint',
    'int64_to_bytes',
    'longlongint_to_bytes',
    'llint_to_bytes',
    'bytes_to_int64',
    'bytes_to_longlongint',
    'bytes_to_llint',
    'int32_to_bytes',
    'int_to_bytes',
    'bytes_to_int32',
    'bytes_to_int',
    'int16_to_bytes',
    'short_to_bytes',
    'bytes_to_int16',
    'bytes_to_short',
    'int8_to_bytes',
    'byte_to_bytes',
    'bytes_to_int8',
    'bytes_to_byte',
    'uint64_to_bytes',
    'ulonglongint_to_bytes',
    'ullint_to_bytes',
    'bytes_to_uint64',
    'bytes_to_ulonglongint',
    'bytes_to_ullint',
    'uint32_to_bytes',
    'uint_to_bytes',
    'bytes_to_uint32',
    'bytes_to_uint',
    'uint16_to_bytes',
    'ushort_to_bytes',
    'bytes_to_uint16',
    'bytes_to_ushort',
    'uint8_to_bytes',
    'ubyte_to_bytes',
    'bytes_to_uint8',
    'bytes_to_ubyte',
    'float64_to_bytes',
    'double_to_bytes',
    'bytes_to_float64',
    'bytes_to_double',
    'float32_to_bytes',
    'float_to_bytes',
    'bytes_to_float32',
    'bytes_to_float',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import struct


# Arbitrary length integers 


def bigint_to_bytes(int_data: int, byteorder: str = 'little'):
    return int_data.to_bytes((int_data.bit_length() + 8) // 8, byteorder, signed=True)  # Add 7 to handle cases where bit_length is not a multiple of 8. Add 1 to ensure that the number is always has a sign bit


bint_to_bytes = bigint_to_bytes


def bytes_to_bigint(bytes_data: bytes, byteorder: str = 'little') -> int:
    return int.from_bytes(bytes_data, byteorder=byteorder, signed=True)


bytes_to_bint = bytes_to_bigint


def ubigint_to_bytes(int_data: int, byteorder: str = 'little') -> bytes:
    return int_data.to_bytes((int_data.bit_length() + 7) // 8, byteorder)  # Add 7 to handle cases where bit_length is not a multiple of 8


ubint_to_bytes = ubigint_to_bytes


def bytes_to_ubigint(bytes_data: bytes, byteorder: str = 'little') -> int:
    return int.from_bytes(bytes_data, byteorder)


bytes_to_ubint = bytes_to_ubigint


# Fixed length integers


def int64_to_bytes(int_data: int)->bytes:
    """
    For a 64 bit signed int in little endian
    :param int_data:
    :return: bytes(); len == 8
    """
    result = struct.pack('<q', int_data)
    return result


longlongint_to_bytes = int64_to_bytes
llint_to_bytes = int64_to_bytes


def bytes_to_int64(bytes_data: bytes)->int:
    """
    For a 64 bit signed int in little endian
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<q', bytes_data)[0]
    return result


bytes_to_longlongint = bytes_to_int64
bytes_to_llint = bytes_to_int64


def int32_to_bytes(int_data: int)->bytes:
    """
    For a 32 bit signed int in little endian
    :param int_data:
    :return: bytes(); len == 4
    """
    result = struct.pack('<i', int_data)
    return result


int_to_bytes = int32_to_bytes


def bytes_to_int32(bytes_data: bytes)->int:
    """
    For a 32 bit signed int in little endian
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<i', bytes_data)[0]
    return result


bytes_to_int = bytes_to_int32


def int16_to_bytes(short_data: int)->bytes:
    """
    For a 16 bit signed short in little endian
    :param short_data:
    :return: bytes(); len == 2
    """
    result = struct.pack('<h', short_data)
    return result


short_to_bytes = int16_to_bytes


def bytes_to_int16(bytes_data: bytes)->int:
    """
    For a 16 bit signed short in little endian
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<h', bytes_data)[0]
    return result


bytes_to_short = bytes_to_int16


def int8_to_bytes(byte_data: int)->bytes:
    """
    For a 8 bit signed byte
    :param byte_data:
    :return: bytes(); len == 1
    """
    result = struct.pack('<b', byte_data)
    return result


byte_to_bytes = int8_to_bytes


def bytes_to_int8(bytes_data: bytes)->int:
    """
    For a 8 bit signed byte
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<b', bytes_data)[0]
    return result


bytes_to_byte = bytes_to_int8


def uint64_to_bytes(int_data: int)->bytes:
    """
    For a 64 bit unsigned int in little endian
    :param int_data:
    :return: bytes(); len == 8
    """
    result = struct.pack('<Q', int_data)
    return result


ulonglongint_to_bytes = uint64_to_bytes
ullint_to_bytes = uint64_to_bytes


def bytes_to_uint64(bytes_data: bytes)->int:
    """
    For a 64 bit unsigned int in little endian
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<Q', bytes_data)[0]
    return result


bytes_to_ulonglongint = bytes_to_uint64
bytes_to_ullint = bytes_to_uint64


def uint32_to_bytes(int_data: int)->bytes:
    """
    For a 32 bit unsigned int in little endian
    :param int_data:
    :return: bytes(); len == 4
    """
    result = struct.pack('<I', int_data)
    return result


uint_to_bytes = uint32_to_bytes


def bytes_to_uint32(bytes_data: bytes)->int:
    """
    For a 32 bit unsigned int in little endian
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<I', bytes_data)[0]
    return result


bytes_to_uint = bytes_to_uint32


def uint16_to_bytes(short_data: int)->bytes:
    """
    For a 16 bit unsigned short in little endian
    :param short_data:
    :return: bytes(); len == 2
    """
    result = struct.pack('<H', short_data)
    return result


ushort_to_bytes = uint16_to_bytes


def bytes_to_uint16(bytes_data: bytes)->int:
    """
    For a 16 bit unsigned short in little endian
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<H', bytes_data)[0]
    return result


bytes_to_ushort = bytes_to_uint16


def uint8_to_bytes(byte_data: int)->bytes:
    """
    For a 8 bit unsigned byte
    :param byte_data:
    :return: bytes(); len == 1
    """
    result = struct.pack('<B', byte_data)
    return result


ubyte_to_bytes = uint8_to_bytes


def bytes_to_uint8(bytes_data: bytes)->int:
    """
    For a 8 bit unsigned byte
    :param bytes_data:
    :return:
    """
    result = struct.unpack('<B', bytes_data)[0]
    return result


bytes_to_ubyte = bytes_to_uint8


# Fixed length floats


def float64_to_bytes(float_data):
    """
    For a 64 bit signed ieee 754 floating points
    http://stackoverflow.com/questions/6286033/reading-32-bit-signed-ieee-754-floating-points-from-a-binary-file-with-python
    :param float_data:
    :return: bytes() with len == 8
    """
    result = struct.pack('d', float_data)
    return result


double_to_bytes = float64_to_bytes


def bytes_to_float64(bytes_data):
    """
    For a 64 bit signed ieee 754 floating points
    http://stackoverflow.com/questions/6286033/reading-32-bit-signed-ieee-754-floating-points-from-a-binary-file-with-python
    :param bytes_data: bytes() with len == 8
    :return:
    """
    result = struct.unpack('d', bytes_data)[0]
    return result


bytes_to_double = bytes_to_float64


def float32_to_bytes(float_data):
    """
    For a 32 bit signed ieee 754 floating points
    http://stackoverflow.com/questions/6286033/reading-32-bit-signed-ieee-754-floating-points-from-a-binary-file-with-python
    :param float_data:
    :return: bytes() with len == 4
    """
    result = struct.pack('f', float_data)
    return result


float_to_bytes = float32_to_bytes


def bytes_to_float32(bytes_data):
    """
    For a 32 bit signed ieee 754 floating points
    http://stackoverflow.com/questions/6286033/reading-32-bit-signed-ieee-754-floating-points-from-a-binary-file-with-python
    :param bytes_data: bytes() with len == 4
    :return:
    """
    result = struct.unpack('f', bytes_data)[0]
    return result


bytes_to_float = bytes_to_float32



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['main']

import sys
import os

# from cengal.bulk_pip_actions.install import _change_current_dir

# import set_environment_variables
#
# set_environment_variables.main()
from cengal.bulk_pip_actions.install import *
from cengal.bulk_pip_actions.bulk_install import *
from typing import List



__author__ = 'ButenkoMS <gtalk@butenkoms.space>'


class PyPiModules(ModulesLists):
    def __init__(self):
        super(PyPiModules, self).__init__()

        self.list_type = FROM_PYPI

        self.python2 = [
            'paver',  # needed for python-lzf module installation process
            'common-mimetypes',  # for pypy2 and Python2 only
            'python-lzf',  # for pypy2 and Python2 only
            # 'line-profiler',  # for CPython only! (as far as I know)
            # 'macropy',

            # ==================
            # OPTIONAL PACKAGES:
        ]

        self.cpython2 = [
            # 'macropy',

            # ==================
            # OPTIONAL PACKAGES:
        ]

        self.cpython3 = [
            # 'macropy3',

            # ==================
            # OPTIONAL PACKAGES:
        ]

        self.pypy2 = [
            # 'macropy',

            # ==================
            # OPTIONAL PACKAGES:
            'msgpack-pypy',
        ]

        self.pypy3 = [
            # 'macropy3',

            # ==================
            # OPTIONAL PACKAGES:
            'msgpack-pypy',
        ]

        self.universal = [
            # 'wheel',
            # 'virtualenv',

            'Cython>=0.29.34',
            'python-dateutil',
            # 'holidays',
            # 'bdateutil',  # Used by Cengal. However requires outdated version of python-dateutil==2.2. Leads to conflicts with 'pendulum' and 'pandas'. Must not be installed. Latest dev version will be installed from github in ExternalGitModules.

            # 'paver',  # needed for python-lzf module installation process (also for zip-module)
            # 'urllib3',  # already preinstalled in Ubuntu 14.04 x64
            # 'thrift',
            # 'thriftpy',
            # 'tornado',
            'requests',  # already preinstalled in Ubuntu 14.04 x64
            # 'pyOpenSSL',
            # 'pycurl',  # already preinstalled in Ubuntu 14.04 x64
            # 'pycrypto',  # outdated. Would not install on Windows easily
            # 'http-parser',
            # 'html5lib',  # already preinstalled in Ubuntu 14.04 x64
            'greenlet',
            # 'cryptography',  # already preinstalled in Ubuntu 14.04 x64
            # 'SQLAlchemy',

            # 'audioread',
            # 'beautifulsoup4',
            # 'cchardet',
            # 'charset_normalizer',
            # 'cookies',
            # 'decorator',
            # 'httpagentparser',
            # 'pyaudio',  # a bit outdated. Would not install on Windows easily
            # 'pygeoip',
            # 'pylast',
            # 'python-mimeparse',
            'colorama',
            # 'bintrees',
            # 'python-Levenshtein',
            # 'httptools',  # Used in tests and benchmarks
            # 'six',

            # 'psycopg2',

            # 'pyinstaller',
            # 'pywin32',  # Would not install on Windows easily. MSI package should be used

            'py-cpuinfo',
            # 'pypreprocessor',
            # 'conditional',

            # 'rocksdb',
            'lmdb',

            # 'json2html',
            # 'json2table',

            # 'Flask',
            # 'MarkupSafe',
            # 'itsdangerous',
            # 'Jinja2',

            'async_generator',
            # 'dependable',  # just interesting package. Was found while searching PyPi for an async_generator package
            # 'janus',
            
            # 'distex',
            
            # 'version',  # https://github.com/keleshev/version Interesting. Python 3 unsupported however. Fixed version from https://github.com/YouCannotBurnMyShadow/version.git should be used instead

            # ==================
            # OPTIONAL PACKAGES:
            'simplejson',
            # 'pyasn1',  # already preinstalled in Ubuntu 14.04 x64
            'cbor',
            'cbor2',
            'ujson',
            'orjson',
            'cloudpickle',
            'python3-dtls',
            'psutil',
            'ttkbootstrap',
            'pprintpp'
        ]

        # self.windows_allowed = {
        #     # 'pyinstaller',
        #     # 'pywin32',
        # }

        self.windows_forbidden = {
            'virtualenv',
            'rocksdb',
            'distex',
        }

        self.emscripten_forbidden = {
            'virtualenv',
            'rocksdb',
            'distex',
            'requests',  # already preinstalled in Ubuntu 14.04 x64
            'greenlet',
            'lmdb',
            'ttkbootstrap'
        }

        if (PYTHON_VERSION_INT[:3] < (3, 4, 0)):
            self.universal.insert(0, 'enum34')  # for pypy3 and Python2 only: it is backport from python34 which is
            #   default in Ubuntu 14.04
        
        if ('CPython'.lower() == PLATFORM_NAME.lower()) and (PYTHON_VERSION_INT[:3] < (3, 3, 0)):
            self.universal.insert(0, 'lzmaffi')  # for pypy3 and Python2 only: it is backport from python34 which is
            #   default in Ubuntu 14.04
        
        if ('PyPy'.lower() == PLATFORM_NAME.lower()) and (PYTHON_VERSION_INT[:3] < (3, 4, 0)):
            self.universal.insert(0, 'lzmaffi')  # for pypy3 and Python2 only: it is backport from python34 which is
            #   default in Ubuntu 14.04
        
        if 'PyPy'.lower() != PLATFORM_NAME.lower():
            self.osx_allowed.add('pyobjc')  # 2024.03.23: PyObjC is not supported on PyPy

        if (PYTHON_VERSION_INT[:3] < (3, 5, 0)):
            self.universal.insert(0, 'typing')  # for pypy3 and Python2 only: it is backport from python35
        
        if (PYTHON_VERSION_INT[:3] < (3, 8, 0)):
            self.cpython2.insert(0, 'msgpack-python')
            self.cpython3.insert(0, 'msgpack-python')
        
        if (PYTHON_VERSION_INT[:3] >= (3, 8, 0)):
            self.cpython2.insert(0, 'msgpack')
            self.cpython3.insert(0, 'msgpack')
        
        if (PYTHON_VERSION_INT[:3] < (3, 11, 1)):
            self.universal.insert(0, 'cchardet')  # 2023.01.27: 3.11.1 is not supported yet: error: longintrepr.h: No such file or directory
            self.universal.insert(0, 'http-parser')  # 2023.01.27: 3.11.1 is not supported yet: error: longintrepr.h: No such file or directory
        
        if (PYTHON_VERSION_INT[:3] < (3, 13, 0)):
            self.universal.insert(0, 'charset_normalizer')  # 2023.11.16: 3.12 is declared to be supported. Testing required
        
        if (PYTHON_VERSION_INT[:3] >= (3, 12, 0)):
            self.windows_allowed.add('patch-ng')  # `Exception: Building py-lmdb from source on Windows requires the "patch-ng" python module.`. 2023.11.18: py-lmdb does not support 3.12 yet. Testing required
            self._forbidden.add('lmdb')  # 2023.11.21: There is no wheels for pypy3.12 and above.
        
        if 'PyPy'.lower() == PLATFORM_NAME.lower():
            if (PYTHON_VERSION_INT[:2] == (3, 10)):
                self.windows_forbidden.add('lmdb')  # 2023.11.21: There is no wheels for pypy3.9 on Windows.

            if (PYTHON_VERSION_INT[:2] != (3, 9)) and (PYTHON_VERSION_INT[:2] != (2, 7)):
                self.windows_allowed.add('patch-ng')  # 2023.11.21: There is only wheels for pypy2.7 and pypy3.9 on Windows

            if (PYTHON_VERSION_INT[:2] != (2, 7)):
                self.osx_allowed.add('patch-ng')  # 2023.11.21: There is only wheel for pypy2.7 on OSX
            
            self._forbidden.add('orjson')  # 2023.11.21: There is no wheels for PyPy on Windows. And for build it requires Rust build tools to be installed
            self.arch__ARM__allowed.add('patch-ng')  # 2023.11.21: There is no prebuilt wheel for ARM
            self.windows_forbidden.add('psutil')  # 2023.11.21: It won't compile under PyPy3.9 or PyPy3.10 on Windows


class ExternalGitModules(ModulesLists):
    def __init__(self):
        super(ExternalGitModules, self).__init__()

        self.list_type = FROM_EXTERNAL_GIT

        self.universal = [
            # 'git+https://github.com/Epikem/pypreprocessor.git',  # temporary workaround for pypreprocessor: https://github.com/interpreters/pypreprocessor/pull/15 . https://github.com/YouCannotBurnMyShadow/pypreprocessor.git must be used instead:
            # 'git+https://github.com/YouCannotBurnMyShadow/pypreprocessor.git',
            'git+https://github.com/YouCannotBurnMyShadow/python-bdateutil.git',  # TODO: check is it really will be installed
            'git+https://github.com/YouCannotBurnMyShadow/version.git',  # TODO: check is it really will be installed
            'git+https://github.com/FI-Mihej/fi-patched--progress.git',  # TODO: check is it really will be installed
        ]

        self.arch__x86_64__allowed = {
        }

        self.arch__x86_32__allowed = {
        }

        self.emscripten_forbidden = {
            'git+https://github.com/FI-Mihej/fi-patched--progress.git',
        }



class ZipModules(ModulesLists):
    def __init__(self):
        super(ZipModules, self).__init__()

        self.list_type = FROM_ZIP


class FoldersWithZippedModules(ModulesLists):
    def __init__(self):
        super(FoldersWithZippedModules, self).__init__()

        self.list_type = FROM_FOLDER_WITH_ZIP_FILES

        dir_of_the_current_file = os.path.dirname(os.path.abspath(__file__))

        self.python3 = [
            os.path.join(dir_of_the_current_file, 'InstallSources', 'Python', 'Packets', 'Python3')
        ]

        self.universal = [
            os.path.join(dir_of_the_current_file, 'InstallSources', 'Python', 'Packets', 'Universal')
        ]


def install_bundled(args=None):
    if args is None:
        args = sys.argv[1:]

    all_modules = [
        ExternalGitModules(),
        ZipModules(),
        FoldersWithZippedModules(),
    ]

    for modules_bunch in all_modules:
        modules_bunch.bulk_install()

    return 0


def get_pypi_requirements_list() -> List[str]:
    pypi_modules: PyPiModules = PyPiModules()
    return pypi_modules.chosen_packages()


def get_remote_requirements_list() -> List[str]:
    result: List[str] = list()
    pypi_modules: PyPiModules = PyPiModules()
    result.extend(pypi_modules.chosen_packages())
    external_git_modules = ExternalGitModules()
    result.extend(external_git_modules.chosen_packages())
    return result


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    all_modules: List[ModulesLists] = [
        PyPiModules(),
        ExternalGitModules(),
        ZipModules(),
        FoldersWithZippedModules(),
    ]

    # path_to_interpreter = os.path.dirname(sys.executable)
    # sys.path.append(path_to_interpreter)
    # with _change_current_dir(path_to_interpreter):
    #     for modules_bunch in all_modules:
    #         modules_bunch.bulk_install()
    for modules_bunch in all_modules:
        modules_bunch.bulk_install()

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['AppDirPath']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# More info:
# https://developer.apple.com/documentation/foundation/1414224-nssearchpathfordirectoriesindoma
# https://developer.apple.com/documentation/foundation/filemanager/searchpathdomainmask
# https://developer.apple.com/documentation/foundation/filemanager/searchpathdirectory


from cengal.file_system.directory_manager import ensure_dir as ensure_dir_exists
from .app_directory_types import AppDirectoryType
from .app_dir_path_base import AppDirPathBase, DirTypeMappingItem, DirTypeMappingItem, BaseDirID, DirNameOrPath, norm_dir_name_or_path
from typing import Optional, Tuple, Dict
from functools import lru_cache
from enum import IntEnum
from Foundation import NSSearchPathForDirectoriesInDomains
from Foundation import NSDocumentDirectory, NSUserDomainMask


class FileManagerSearchPathDirectory(IntEnum):
    applicationDirectory = 1
    userDirectory = 7
    cachesDirectory = 13
    applicationSupportDirectory = 14
    allApplicationsDirectory = 100


class FileManagerSearchPathDomainMask(IntEnum):
    userDomainMask = 1
    localDomainMask = 4
    systemDomainMask = 8
    networkDomainMask = 16
    allDomainsMask = 65535


dir_type_mapping: Dict[AppDirectoryType, Tuple[Tuple[FileManagerSearchPathDomainMask, FileManagerSearchPathDirectory], Optional[DirNameOrPath], Optional[DirNameOrPath]]] = {
    AppDirectoryType.local_data: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local', 'data')),
    AppDirectoryType.local_static_data: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local', 'static_data')),
    AppDirectoryType.local_cache: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.cachesDirectory), None, ('local', 'cache')),
    AppDirectoryType.local_config: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local', 'config')),
    AppDirectoryType.local_log: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local', 'log')),
    AppDirectoryType.local_temp: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.cachesDirectory), None, ('local', 'temp')),
    AppDirectoryType.local_runtime: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local', 'runtime')),

    AppDirectoryType.local_low_data: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local_low', 'data')),
    AppDirectoryType.local_low_static_data: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local_low', 'static_data')),
    AppDirectoryType.local_low_cache: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.cachesDirectory), None, ('local_low', 'cache')),
    AppDirectoryType.local_low_config: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local_low', 'config')),
    AppDirectoryType.local_low_log: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local_low', 'log')),
    AppDirectoryType.local_low_temp: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.cachesDirectory), None, ('local_low', 'temp')),
    AppDirectoryType.local_low_runtime: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('local_low', 'runtime')),

    AppDirectoryType.roaming_data: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('roaming', 'data')),
    AppDirectoryType.roaming_static_data: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('roaming', 'static_data')),
    AppDirectoryType.roaming_cache: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.cachesDirectory), None, ('roaming', 'cache')),
    AppDirectoryType.roaming_config: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('roaming', 'config')),
    AppDirectoryType.roaming_log: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('roaming', 'log')),
    AppDirectoryType.roaming_temp: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.cachesDirectory), None, ('roaming', 'temp')),
    AppDirectoryType.roaming_runtime: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, ('roaming', 'runtime')),

    AppDirectoryType.program_data: ((FileManagerSearchPathDomainMask.localDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, 'data'),
    AppDirectoryType.program_static_data: ((FileManagerSearchPathDomainMask.localDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, 'static_data'),
    AppDirectoryType.program_cache: ('/Library/Caches', None, 'cache'),
    AppDirectoryType.program_config: ((FileManagerSearchPathDomainMask.localDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, 'config'),
    AppDirectoryType.program_log: ((FileManagerSearchPathDomainMask.localDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, 'log'),
    AppDirectoryType.program_temp: ('/tmp', None, 'temp'),
    AppDirectoryType.program_runtime: ((FileManagerSearchPathDomainMask.localDomainMask, FileManagerSearchPathDirectory.applicationSupportDirectory), None, 'runtime'),

    AppDirectoryType.program_files: ((FileManagerSearchPathDomainMask.localDomainMask, FileManagerSearchPathDirectory.allApplicationsDirectory), None, None),
    AppDirectoryType.program_files_common: ((FileManagerSearchPathDomainMask.localDomainMask, FileManagerSearchPathDirectory.allApplicationsDirectory), None, None),

    AppDirectoryType.user_profile_data: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.userDirectory), None, None),
    AppDirectoryType.user_profile_program_files: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationDirectory), None, None),
    AppDirectoryType.user_profile_program_files_common: ((FileManagerSearchPathDomainMask.userDomainMask, FileManagerSearchPathDirectory.applicationDirectory), None, None),
}


class AppDirPath(AppDirPathBase):
    def __init__(self, max_cache_size: Optional[int] = None) -> None:
        super().__init__(max_cache_size)
    
    def dir_type_mapping(self, dir_type: AppDirectoryType) -> DirTypeMappingItem:
        return dir_type_mapping[dir_type]
    
    def base_dir_id_to_path(self, base_dir_id: BaseDirID) -> str:
        if isinstance(base_dir_id, str):
            return base_dir_id
        
        domain_mask, directory = base_dir_id
        result_path = NSSearchPathForDirectoriesInDomains(directory.value, domain_mask.value, True)
        if result_path:
            # Typically, you're interested in the first result
            return result_path[0]
        else:
            raise RuntimeError('NSSearchPathForDirectoriesInDomains returned None for domain_mask: {0}, directory: {1}'.format(domain_mask, directory))



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'get_file_hash', 
    'get_file_modification_date', 
    'current_src_file_dir', 
    'path_relative_to_current_src', 
    'path_relative_to_current_dir', 
    'file_exists', 
    'full_ext', 
    'full_ext_parts', 
    'last_ext', 
    'has_ext', 
    'file_name', 
]


import hashlib
# import os
import datetime
import inspect
from typing import Optional
from os import getcwd, extsep
from os.path import exists, isfile, normpath, dirname, getmtime, realpath, join, basename

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def get_file_hash(full_file_name, hash_format_string=None):
    hash_format_string = hash_format_string or '{} {}'
    file_content = None
    with open(full_file_name, 'rb') as file:
        file_content = file.read()
    file_hash = hash_format_string.format(hashlib.sha512(file_content).hexdigest(), hex(len(file_content))[2:])
    return file_hash


def get_file_modification_date(full_file_name):
    time_stamp = getmtime(full_file_name)
    return datetime.datetime.fromtimestamp(time_stamp)


def current_src_file_dir() -> str:
    return dirname(realpath(inspect.currentframe().f_back.f_code.co_filename))


def path_relative_to_current_src(relative_path: Optional[str]=None) -> str:
    relative_path = relative_path or str()
    return normpath(join(dirname(realpath(inspect.currentframe().f_back.f_code.co_filename)), normpath(relative_path)))


def path_relative_to_current_dir(relative_path: Optional[str]=None) -> str:
    relative_path = relative_path or str()
    return normpath(join(getcwd(), normpath(relative_path)))


def file_exists(file_path: str) -> bool:
    return exists(file_path) and isfile(file_path)


def full_ext(file_path: str) -> str:
    return extsep.join(basename(normpath(file_path)).split(extsep)[1:])


def full_ext_parts(file_path: str) -> str:
    return basename(normpath(file_path)).split(extsep)[1:]


def last_ext(file_path: str) -> str:
    return basename(normpath(file_path)).split(extsep)[-1]


def has_ext(file_path: str) -> str:
    return extsep in basename(normpath(file_path))


def file_name(file_path: str) -> str:
    return basename(normpath(file_path)).split(extsep)[0]



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.system import PLATFORM_NAME
if 'PyPy' != PLATFORM_NAME:
    import uuid
from enum import Enum

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class TypeOfGenerator(Enum):
    INTEGER = 0
    GUID_STRING = 1

    # _VALUES_TO_NAMES = {
    #     0: "INTEGER",
    #     1: "GUID_STRING",
    # }
    #
    # _NAMES_TO_VALUES = {
    #     "INTEGER": 0,
    #     "GUID_STRING": 1,
    # }


class IDGenerator:

    def __init__(self, type=TypeOfGenerator.INTEGER):
        self.counter = 0
        self.type = type

    def get_new_ID(self):
        currentCounter = None
        if TypeOfGenerator.INTEGER == self.type:
            currentCounter = self.counter
            self.counter += 1
        elif TypeOfGenerator.GUID_STRING == self.type:
            seq = self.counter
            self.counter += 1
            currentCounter = uuid.uuid1(clock_seq=seq).hex
        return currentCounter

    def remove_ID(self, ID):
        pass

    def clear(self):
        self.counter = 0

    def __call__(self):
        return self.get_new_ID()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['PointBase', 'PointDimensionIndexError', 'PointNd', 'Point1d', 'Point2d', 'Point3d', 'PointNdXYZ', 'convert_point_to_xyz']

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from copy import copy
from cengal.entities.copyable import *
from typing import Any, FrozenSet, Optional, Sequence, List, Set, Tuple, Dict, Union
numpy_present = True
try:
    import numpy as np
    from numpy import array_equal, array, ndarray
except:
    numpy_present = False

RawListPointType = List[Union[float, int]]
if numpy_present:
    RawNdarrayPointType = ndarray
    RawUniversalPointType = Union[RawListPointType, RawNdarrayPointType]
else:
    RawUniversalPointType = RawListPointType


class PointBase(CopyableMixin):
    dimension: int = 0

    def copy(self) -> 'PointBase':
        raise NotImplementedError

    def shallow_copy(self) -> 'PointBase':
        raise NotImplementedError
    
    def updated_copy(self, update: Dict) -> 'PointBase':
        raise NotImplementedError

    @property
    def dim(self) -> int:
        return self.dimension

    def __call__(self, *args: Any, **kwds: Any) -> RawListPointType:
        raise NotImplementedError
    
    def clear(self):
        raise NotImplementedError


class PointDimensionIndexError(Exception):
    pass


class PointWrongNdarrayTypeError(Exception):
    pass


class PointNd(PointBase):
    dimension: int = None
    _dim_names_translation_table: Tuple[Tuple[Set, int]] = tuple()
    _default_ndarray_type = float
    _default_shallow_copy = False

    def __init__(self
            , dimension: int
            , *args: Union[Tuple[Sequence[Union[float, int]]], Tuple[Union[float, int]]]
            , **kwargs: Dict[str, Union[float, int]]
            ) -> None:
        shallow_copy = kwargs.get('shallow_copy', self._default_shallow_copy)
        kwargs.pop('shallow_copy', None)
        if shallow_copy and args and isinstance(args[0], PointNd):
            first_element = args[0]
            self.dimension = first_element.dimension
            self._point = first_element._point
            self._ndarray_type = first_element._ndarray_type
        else:
            if dimension is not None:
                self.dimension: int = dimension
            
            self._point: RawListPointType = None
            self._ndarray_type = kwargs.get('ndarray_type', self._default_ndarray_type)
            kwargs.pop('ndarray_type', None)
            if self._ndarray_type not in {int, float}:
                print(f'Wrong type: {self._ndarray_type}')
                raise PointWrongNdarrayTypeError
            
            self.clear()
            self(*args, **kwargs)
    
    def __repr__(self):
        return f'{type(self).__name__}(dimension={self.dimension}, point=({", ".join((str(item) for item in self._point))}), ndarray_type={self._ndarray_type})'
    
    def copy(self) -> 'PointNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['dimension'] = self.dimension
        result.__dict__['_point'] = self._point.copy()
        result.__dict__['_ndarray_type'] = self._ndarray_type
        return result
    
    def shallow_copy(self) -> 'PointNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['dimension'] = self.dimension
        result.__dict__['_point'] = self._point
        result.__dict__['_ndarray_type'] = self._ndarray_type
        return result
    
    def updated_copy(self, update: Dict) -> 'PointNd':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__['dimension'] = self.dimension
        result.__dict__['_point'] = get_dict_key_with_callable_default(update, '_point', lambda: self._point.copy())
        result.__dict__['_ndarray_type'] = self._ndarray_type
        return result
    
    def clear(self):
        self._point = [self._ndarray_type()] * self.dim
        if numpy_present:
            self._point = array(self._point)

    def __call__(self
            , *args: Union[Tuple[Sequence[Union[float, int]]], Tuple[Union[float, int]]]
            , **kwargs: Dict[str, Union[float, int]]
            ) -> RawUniversalPointType:
        if args:
            first_element = args[0]
            data = first_element
            if isinstance(first_element, PointNd):
                data = first_element()
            elif isinstance(first_element, (float, int)):
                data = args

            for index, item in enumerate(data):
                if index >= self.dim:
                    break

                self._point[index] = item
        elif kwargs:
            for dim_name_variants, dim_index in self._dim_names_translation_table:
                for dim_name in dim_name_variants:
                    if dim_name in kwargs:
                        kwargs[f'n{dim_index}'] = kwargs[dim_name]
                        break

            for dim_index in range(self.dimension):
                dim_name = f'n{dim_index}'
                dim_value = kwargs.get(dim_name, None)
                if dim_value is not None:
                    self._point[dim_index] = dim_value
        
        return self._point
    
    def as_list(self) -> RawListPointType:
        return list(self._point)
    
    def __setattr__(self, __name: str, __value: Any) -> None:
        for dim_name_variants, dim_index in self._dim_names_translation_table:
            if __name in dim_name_variants:
                self._point[dim_index] = __value
                return __value

        if __name and __name.startswith('n'):
            index_str = __name[1:]
            if index_str.isnumeric():
                index = int(index_str)
                if index < self.dim:
                    self._point[index] = __value
                    return __value
                else:
                    raise PointDimensionIndexError

        return super().__setattr__(__name, __value)
    
    def __getattribute__(self, __name: str) -> Any:
        for dim_name_variants, dim_index in super().__getattribute__('_dim_names_translation_table'):
            if __name in dim_name_variants:
                return super().__getattribute__('_point')[dim_index]

        if __name and __name.startswith('n'):
            index_str = __name[1:]
            if index_str.isnumeric():
                index = int(index_str)
                if index < super().__getattribute__('dim'):
                    return super().__getattribute__('_point')[index]
                else:
                    raise PointDimensionIndexError

        return super().__getattribute__(__name)
    
    def __getitem__(self, index_or_key):
        if isinstance(index_or_key, int):
            return self._point[index_or_key]
        elif isinstance(index_or_key, str):
            for dim_name_variants, dim_index in self._dim_names_translation_table:
                if index_or_key in dim_name_variants:
                    return self._point[dim_index]

            if index_or_key and index_or_key.startswith('n'):
                index_str = index_or_key[1:]
                if index_str.isnumeric():
                    index = int(index_str)
                    if index < self.dim:
                        return self._point[index]
                    else:
                        raise PointDimensionIndexError
        else:
            raise KeyError

    def __setitem__(self, index_or_key, value):
        if isinstance(index_or_key, int):
            self._point[index_or_key] = value
        elif isinstance(index_or_key, str):
            for dim_name_variants, dim_index in self._dim_names_translation_table:
                if index_or_key in dim_name_variants:
                    self._point[dim_index] = value

            if index_or_key and index_or_key.startswith('n'):
                index_str = index_or_key[1:]
                if index_str.isnumeric():
                    index = int(index_str)
                    if index < self.dim:
                        self._point[index] = value
                    else:
                        raise PointDimensionIndexError
        else:
            raise KeyError
    
    def __len__(self):
        return self.dim

    def crop_to(self, *args) -> 'PointNd':
        """Example:
        >> p4 = Point3d(3, 5, 2, 8)
        >> p3 = convert_point_to_xyz(p4.crop_to('z', 'n0', 3))
        >> print(p3)
        Point3d(2, 3, 8)

        Raises:
            PointDimensionIndexError: _description_
            KeyError: _description_

        Returns:
            PointNd: _description_
        """
        if args:
            first_item = args[0]
            data = first_item
            if isinstance(first_item, (int, str)):
                data = args
            
            index_sequence = list()
            for index_or_key in data:
                if isinstance(index_or_key, int):
                    index_sequence.append(index_or_key)
                elif isinstance(index_or_key, str):
                    for dim_name_variants, dim_index in self._dim_names_translation_table:
                        if index_or_key in dim_name_variants:
                            index_sequence.append(dim_index)

                    if index_or_key and index_or_key.startswith('n'):
                        index_str = index_or_key[1:]
                        if index_str.isnumeric():
                            index = int(index_str)
                            if index < self.dim:
                                index_sequence.append(index)
                            else:
                                raise PointDimensionIndexError
                else:
                    raise KeyError
            
            croped_data = list()
            for index in index_sequence:
                croped_data.append(self._point[index])
            
            return PointNd(len(croped_data), croped_data)
        else:
            return self.copy()

    def expand_to(self, dimension: int, *args) -> 'PointNd':
        """Example:
        >> p2 = Point2d(2, 3)
        >> p3 = convert_point_to_xyz(p2.expand_to(3, 2, 'n0'))
        >> print(p3)
        Point3d(3, 0, 2)

        Args:
            dimension (int): _description_
            *args (Union[Tuple[int], Tuple[str]]): list used to map `point item` with index equal to `list's item index` to `result point item` with index equals to `list's item value`. If value is str: it should starts from 'n' char and continue with (positive) numeric chars only

        Raises:
            PointDimensionIndexError: when `len(args) > len(self)`
            KeyError: wrong type of *args or type/format of it's value(s)

        Returns:
            PointNd: _description_
        """
        if args:
            first_item = args[0]
            data = first_item
            if isinstance(first_item, (int, str)):
                data = args
            
            index_sequence = list()
            for index_or_key in data:
                if isinstance(index_or_key, int):
                    index_sequence.append(index_or_key)
                elif isinstance(index_or_key, str):
                    if index_or_key and index_or_key.startswith('n'):
                        index_str = index_or_key[1:]
                        if index_str.isnumeric():
                            index = int(index_str)
                            if index < dimension:
                                index_sequence.append(index)
                            else:
                                raise PointDimensionIndexError
                        else:
                            raise KeyError
                    else:
                        raise KeyError
                else:
                    raise KeyError
            
            expansion_data = dict()
            for source_index, destination_index in enumerate(index_sequence):
                expansion_data[f'n{destination_index}'] = self._point[source_index]
            
            return PointNd(dimension, **expansion_data)
        else:
            return self.copy()
    
    def __eq__(self, other) -> bool:
        if isinstance(other, PointNd):
            if numpy_present:
                return (self.dim == other.dim) and array_equal(self._point, other._point)
            else:
                return (self.dim == other.dim) and (self._point == other._point)
        
        return False
    
    def __ne__(self, other) -> bool:
        return not self.__eq__(other)
    
    def __hash__(self) -> int:
        return hash((self.dim, tuple(self._point)))
    
    def __bool__(self) -> bool:
        """If not zero point

        Returns:
            bool: _description_
        """
        for item in self._point:
            if item:
                return True
        
        return False


class Point1d(PointNd):
    dimension: int = 1
    _dim_names_translation_table: Tuple[Tuple[Set, int]] = (
        ({'x', 'X'}, 0),
    )

    def __init__(self
            , *args: Union[Tuple[Sequence[Union[float, int]]], Tuple[Union[float, int]]]
            , **kwargs: Dict[str, Union[float, int]]
            ) -> None:
        PointNd.__init__(self, None, *args, **kwargs)


class Point2d(PointNd):
    dimension: int = 2
    _dim_names_translation_table: Tuple[Tuple[Set, int]] = (
        ({'x', 'X'}, 0),
        ({'y', 'Y'}, 1),
    )

    def __init__(self
            , *args: Union[Tuple[Sequence[Union[float, int]]], Tuple[Union[float, int]]]
            , **kwargs: Dict[str, Union[float, int]]
            ) -> None:
        PointNd.__init__(self, None, *args, **kwargs)


class Point3d(PointNd):
    dimension: int = 3
    _dim_names_translation_table: Tuple[Tuple[Set, int]] = (
        ({'x', 'X'}, 0),
        ({'y', 'Y'}, 1),
        ({'z', 'Z'}, 2),
    )

    def __init__(self
            , *args: Union[Tuple[Sequence[Union[float, int]]], Tuple[Union[float, int]]]
            , **kwargs: Dict[str, Union[float, int]]
            ) -> None:
        PointNd.__init__(self, None, *args, **kwargs)


class PointNdXYZ(PointNd):
    dimension: int = None
    _dim_names_translation_table: Tuple[Tuple[Set, int]] = (
        ({'x', 'X'}, 0),
        ({'y', 'Y'}, 1),
        ({'z', 'Z'}, 2),
    )

    def __init__(self
            , dimension: int
            , *args: Union[Tuple[Sequence[Union[float, int]]], Tuple[Union[float, int]]]
            , **kwargs: Dict[str, Union[float, int]]
            ) -> None:
        super().__init__(dimension, *args, **kwargs)


def convert_point_to_xyz(point: PointNd) -> PointNd:
    if 1 == point.dim:
        return Point1d(point, shallow_copy=True)
    elif 2 == point.dim:
        return Point2d(point, shallow_copy=True)
    elif 3 == point.dim:
        return Point3d(point, shallow_copy=True)
    elif 3 < point.dim:
        return PointNdXYZ(point, shallow_copy=True)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['CpuInfo', 'cpu_info']


# from operator import is_
from cengal.modules_management.alternative_import import alt_import
with alt_import('cpuinfo') as cpuinfo:
    if cpuinfo is None:
        CPUINFO_PRESENT: bool = False
    else:
        CPUINFO_PRESENT = True

from cengal.modules_management.ignore_in_build_mode import ignore_in_build_mode
PSUTIL_PRESENT: bool = False
with ignore_in_build_mode():
    with alt_import('psutil') as psutil:
        if psutil is not None:
            PSUTIL_PRESENT = True

import sys
from typing import Dict, Set, Optional


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


if hasattr(cpuinfo.cpuinfo, '_check_arch'):
    check_arch = getattr(cpuinfo.cpuinfo, '_check_arch')
    try:
        check_arch()
    except:
        CPUINFO_PRESENT = False


_known_gathering_methods_names = (
    '_get_cpu_info_from_wmic',
    '_get_cpu_info_from_registry',
    '_get_cpu_info_from_proc_cpuinfo',
    '_get_cpu_info_from_cpufreq_info',
    '_get_cpu_info_from_lscpu',
    '_get_cpu_info_from_sysctl',
    '_get_cpu_info_from_kstat',
    '_get_cpu_info_from_dmesg',
    '_get_cpu_info_from_cat_var_run_dmesg_boot',
    '_get_cpu_info_from_ibm_pa_features',
    '_get_cpu_info_from_sysinfo',
    '_get_cpu_info_from_cpuid',
    '_get_cpu_info_from_platform_uname',
)
last_stage = len(_known_gathering_methods_names) - 1
stage_after_last = last_stage + 1


_can_use_lazy_gathering = True
if hasattr(cpuinfo.cpuinfo, '_copy_new_fields'):
    copy_new_fields = getattr(cpuinfo.cpuinfo, '_copy_new_fields')
elif hasattr(cpuinfo.cpuinfo, 'CopyNewFields'):
    copy_new_fields = getattr(cpuinfo.cpuinfo, 'CopyNewFields')
else:
    _can_use_lazy_gathering = False


def get_cpu_info_lazy(desired_keys = None, stage = 0, info = None):
    stage = stage if stage is not None else 0
    stage = stage if stage >= 0 else last_stage

    arch, bits = cpuinfo.cpuinfo._parse_arch(cpuinfo.cpuinfo.DataSource.arch_string_raw)

    friendly_maxsize = { 2**31-1: '32 bit', 2**63-1: '64 bit' }.get(sys.maxsize) or 'unknown bits'
    friendly_version = "{0}.{1}.{2}.{3}.{4}".format(*sys.version_info)
    PYTHON_VERSION = "{0} ({1})".format(friendly_version, friendly_maxsize)

    info = {
        'python_version' : PYTHON_VERSION,
        'cpuinfo_version' : cpuinfo.cpuinfo.CPUINFO_VERSION,
        'cpuinfo_version_string' : cpuinfo.cpuinfo.CPUINFO_VERSION_STRING,
        'arch' : arch,
        'bits' : bits,
        'count' : cpuinfo.cpuinfo.DataSource.cpu_count,
        'arch_string_raw' : cpuinfo.cpuinfo.DataSource.arch_string_raw,
    } if info is None else info

    if desired_keys is None:
        return info, stage

    desired_keys = desired_keys if isinstance(desired_keys, set) else set(desired_keys)
    flags: bool = 'flags' in desired_keys

    if (not flags) & (not (desired_keys - set(info.keys()))):
        return info, stage
    
    current_stage = -1
    for method_name in _known_gathering_methods_names:
        current_stage += 1
        if current_stage < stage:
            continue

        if hasattr(cpuinfo.cpuinfo, method_name):
            method = getattr(cpuinfo.cpuinfo, method_name)
            copy_new_fields(info, method())
            if (not flags) & (not (desired_keys - set(info.keys()))):
                break
    
    next_stage = current_stage + 1
    if next_stage >= stage_after_last:
        next_stage = None
    
    return info, next_stage


class CpuInfo:
    _cache: Dict = None
    _stage: int = None
    _cache_friendly: Dict = None
    _cores_num: int = None
    _virtual_cores_num: int = None

    def __init__(self, desired_keys: Optional[Set[str]] = None, stage: Optional[int] = 0, info: Dict = None):
        if CpuInfo._cores_num is None:
            CpuInfo._cores_num = psutil.cpu_count(logical=False) if PSUTIL_PRESENT else 0
        
        if CpuInfo._virtual_cores_num is None:
            CpuInfo._virtual_cores_num = psutil.cpu_count(logical=True) if PSUTIL_PRESENT else 0
        
        if CpuInfo._cache is None:
            if _can_use_lazy_gathering:
                CpuInfo._cache_friendly, CpuInfo._stage = get_cpu_info_lazy(desired_keys, stage, info) if CPUINFO_PRESENT else ((dict(), stage) if info is None else (info, stage))
            else:
                CpuInfo._cache_friendly = cpuinfo.get_cpu_info() if CPUINFO_PRESENT else (dict() if info is None else info)
            
            CpuInfo._cache = self._normalize_cpu_info_values(CpuInfo._cache_friendly)
    
    def _normalize_cpu_info_values(self, cpu_info_dict: Dict) -> Dict:
        result = dict()
        frequency: Dict[str, int] = {
                'ghz': 10 ** 9, 'mhz': 10 ** 6, 'khz': 10 ** 3, 'hz': 1
            }
        memory_size: Dict[str, int] = {
                'tib': 10 ** 12, 'gib': 10 ** 9, 'mib': 10 ** 6, 'kib': 10 ** 3, 'b': 1,
                'tb': 10 ** 12, 'gb': 10 ** 9, 'mb': 10 ** 6, 'kb': 10 ** 3,
                't': 10 ** 12, 'g': 10 ** 9, 'm': 10 ** 6, 'k': 10 ** 3,
            }
        modif_per_key: Dict[str, Dict[str, int]] = {
            'l1_data_cache_size': memory_size,
            'l1_instruction_cache_size': memory_size,
            'l2_cache_size': memory_size,
            'l3_cache_size': memory_size,
        }
        for key, text_value in cpu_info_dict.items():
            value = text_value
            if isinstance(text_value, str):
                if key in modif_per_key:
                    conversion_required = False
                    try:
                        value = int(text_value)
                    except ValueError:
                        conversion_required = True
                    
                    if conversion_required:
                        modif: Dict[str, int] = modif_per_key[key]
                        is_ok = False
                        minus_data_len = -len(text_value)
                        index = -1
                        while index > minus_data_len:
                            try:
                                value = int(text_value[:index])
                                is_ok = True
                                break
                            except ValueError:
                                pass

                            index -= 1
                        
                        if not is_ok:
                            index = -1
                            while index > minus_data_len:
                                try:
                                    value = float(text_value[:index])
                                    is_ok = True
                                    break
                                except ValueError:
                                    pass

                            index -= 1
                        
                        if is_ok:
                            modif_text: str = text_value[index:]
                            modif_text = modif_text.strip().casefold()
                            modif_value = modif.get(modif_text, 1)
                            value = int(value * modif_value)
                        else:
                            value = 0
                else:
                    try_with_float: bool = False
                    try:
                        value = int(text_value)
                    except ValueError:
                        try_with_float = True
                    
                    if try_with_float:
                        try:
                            value = float(text_value)
                        except ValueError:
                            value = text_value
            
            result[key] = value
        
        return result

    def _ensure_field(self, field_name):
        if not CPUINFO_PRESENT:
            return field_name
        
        if not _can_use_lazy_gathering:
            return field_name

        if ('flags' != field_name) & (field_name in self._cache):
            return field_name

        CpuInfo._cache_friendly, CpuInfo._stage = get_cpu_info_lazy({field_name,}, CpuInfo._stage, CpuInfo._cache_friendly)
        CpuInfo._cache.clear()
        CpuInfo._cache.update(self._normalize_cpu_info_values(CpuInfo._cache_friendly))
        return field_name
    
    @property
    def python_version(self):
        return self._cache.get(self._ensure_field('python_version'), str())
    
    @property
    def cpuinfo_version(self):
        return self._cache.get(self._ensure_field('cpuinfo_version'), (0, 0, 0))
    
    @property
    def cpuinfo_version_string(self):
        return self._cache.get(self._ensure_field('cpuinfo_version_string'), str())
    
    @property
    def python_hz_advertised_friendlyversion(self):
        return self._cache.get(self._ensure_field('hz_advertised_friendly'), str())
    
    @property
    def hz_actual_friendly(self):
        return self._cache.get(self._ensure_field('hz_actual_friendly'), str())
    
    @property
    def hz_advertised(self):
        return self._cache.get(self._ensure_field('hz_advertised'), (0, 0))
    
    @property
    def hz_actual(self):
        return self._cache.get(self._ensure_field('hz_actual'), (0, 0))
    
    @property
    def arch(self):
        return self._cache.get(self._ensure_field('arch'), str())
    
    @property
    def bits(self):
        return self._cache.get(self._ensure_field('bits'), 0)
    
    @property
    def count(self):
        return self._cache.get(self._ensure_field('count'), 0)
    
    @property
    def l1_data_cache_size(self):
        return self._cache.get(self._ensure_field('l1_data_cache_size'), 0)
    
    @property
    def l1_instruction_cache_size(self):
        return self._cache.get(self._ensure_field('l1_instruction_cache_size'), 0)
    
    @property
    def l2_cache_size(self):
        return self._cache.get(self._ensure_field('l2_cache_size'), 0)
    
    @property
    def l2_cache_line_size(self):
        return self._cache.get(self._ensure_field('l2_cache_line_size'), 0)
    
    @property
    def l2_cache_associativity(self):
        return self._cache.get(self._ensure_field('l2_cache_associativity'), 0)
    
    @property
    def l3_cache_size(self):
        return self._cache.get(self._ensure_field('l3_cache_size'), 0)
    
    @property
    def stepping(self):
        return self._cache.get(self._ensure_field('stepping'), 0)
    
    @property
    def model(self):
        return self._cache.get(self._ensure_field('model'), 0)
    
    @property
    def family(self):
        return self._cache.get(self._ensure_field('family'), 0)
    
    @property
    def processor_type(self):
        return self._cache.get(self._ensure_field('processor_type'), 0)
    
    @property
    def flags(self):
        return self._cache.get(self._ensure_field('flags'), list())
    
    @property
    def vendor_id_raw(self):
        return self._cache.get(self._ensure_field('vendor_id_raw'), str())
    
    @property
    def hardware_raw(self):
        return self._cache.get(self._ensure_field('hardware_raw'), str())
    
    @property
    def brand_raw(self):
        return self._cache.get(self._ensure_field('brand_raw'), str())
    
    @property
    def arch_string_raw(self):
        return self._cache.get(self._ensure_field('arch_string_raw'), str())
    
    @property
    def cores_num(self):
        return CpuInfo._cores_num if CpuInfo._cores_num else self.count
    
    @property
    def virtual_cores_num(self):
        return CpuInfo._virtual_cores_num if CpuInfo._virtual_cores_num else CpuInfo._cores_num
    
    @property
    def l2_cache_size_per_core(self) -> int:
        if 0 == self.count:
            return self.l2_cache_size
        
        return int(self.l2_cache_size / self.cores_num)
    
    @property
    def l2_cache_size_per_virtual_core(self) -> int:
        if 0 == self.count:
            return self.l2_cache_size
        
        return int(self.l2_cache_size / self.virtual_cores_num)
    
    @property
    def l3_cache_size_per_core(self) -> int:
        if 0 == self.count:
            return self.l3_cache_size
        
        return int(self.l3_cache_size / self.cores_num)
    
    @property
    def l3_cache_size_per_virtual_core(self) -> int:
        if 0 == self.count:
            return self.l3_cache_size
        
        return int(self.l3_cache_size / self.virtual_cores_num)
    
    @property
    def is_x86(self) -> bool:
        return self.arch.lower().startswith('x86')
    
    @property
    def is_arm(self) -> bool:
        return self.arch.lower().startswith('arm')
    
    def gather_all(self):
        self.flags


_CPU_INFO = None


def cpu_info() -> CpuInfo:
    global _CPU_INFO
    if _CPU_INFO is None:
        _CPU_INFO = CpuInfo()
    
    return _CPU_INFO



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['start_admin', 'cs_init', 'SchedulerPerformanceFormatter', 'scheduler_stats_aggregator_provider', 'CSStatsFormatterMultiprocess',
           'CorosLogsProvider', 'CoroLogsAppendFormatter', 'CoroLogsAppendFormatterParts']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.time_management.cpu_clock_cycles import perf_counter
import ttkbootstrap as ttkb
from ttkbootstrap.scrolled import ScrolledText as TtkbScrolledText
from tkinter import simpledialog
import tkinter as tk
from ttkbootstrap import Style
from pprintpp import pformat as pf
from typing import Callable, Hashable, Optional, Set, Dict, Tuple, Union, List, Sequence, Any, cast

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.tkinter import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro, PutCoroRequest
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import AsyncioLoopRequest, run_in_thread_pool, run_in_thread_pool_fast
from cengal.parallel_execution.coroutines.coro_standard_services.fast_aggregator import *
from cengal.parallel_execution.coroutines.coro_standard_services.log import *
from cengal.parallel_execution.coroutines.coro_standard_services.instance import InstanceRequest
from cengal.parallel_execution.coroutines.coro_tools.coro_flow_control import graceful_coro_destroyer, agraceful_coro_destroyer, GracefulCoroDestroy
from cengal.parallel_execution.coroutines.coro_tools.lock import Lock
from cengal.statistics.normal_distribution import average
from cengal.parallel_execution.multiprocess.multiprocessing_task_runner import *
from cengal.text_processing.text_processing import normalize_line_separators_and_tabs
from cengal.introspection.inspect import get_exception, exception_to_printable_text
from cengal.system import OS_TYPE
from cengal.data_manipulation.dict_path import *
from cengal.user_interface.gui.tkinter.components.read_only_text import *
from cengal.user_interface.gui.tkinter.components.tool_tip import *
from cengal.user_interface.gui.tkinter.components.aggregator_view import *
from cengal.math.numbers import RationalNumber
from cengal.code_flow_control.args_manager import args_kwargs_to_str, ArgsKwargs
from greenlet import GreenletExit
from enum import IntFlag, auto
from uuid import uuid4


command_executor_aggregator_view_key = 'command_executor_aggregator'
command_executor_aggregator_append_view_key = 'command_executor_aggregator_append'

coro_holder_header = 'async def coro_holder(i: Interface, fac: FastAggregatorClient, aggregator_key, aggregator_append_key):'
coro_holder_template = f'''from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro_list import *
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro_list import *
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import *
from cengal.parallel_execution.coroutines.coro_standard_services.throw_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.throw_coro_list import *
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import *
from cengal.parallel_execution.coroutines.coro_standard_services.fast_aggregator import *
from cengal.parallel_execution.coroutines.coro_standard_services.log import *
from cengal.parallel_execution.coroutines.coro_tools.coro_flow_control import graceful_coro_destroyer, GracefulCoroDestroy
import asyncio


{coro_holder_header}
{{coro_body}}


result = i(RunCoro, coro_holder, FastAggregatorClient(), aggregator_key, aggregator_append_key)
'''


def prepare_coro_body(text: str) -> str:
    if (not text) or ('\n' == text):
        text = 'pass'
    
    text = normalize_line_separators_and_tabs(text)
    text_lines = text.splitlines()
    return '\n'.join([' ' * 4 + line for line in text_lines])


class CommandExecutor(ttkb.Frame):
    def __init__(self, width, height, allow_asyncio: bool = True, priority: Optional[CoroPriority] = None, interrupt_when_no_requests: Optional[bool] = None, *args, **kwargs) -> None:
        self.width = width
        self.height = height
        self.allow_asyncio = allow_asyncio
        self.priority = priority
        self.interrupt_when_no_requests = interrupt_when_no_requests
        self.current_coro_id: CoroID = None
        if allow_asyncio:
            i = current_interface()
            i(AsyncioLoopRequest().ensure_loop(interrupt_when_no_requests=interrupt_when_no_requests, priority=priority))
            i(AsyncioLoopRequest().turn_on_loops_intercommunication())

        super().__init__(*args, **kwargs)
        text_area_config = {
            'highlightcolor': Style.instance.colors.primary,
            'highlightbackground': Style.instance.colors.border,
            'highlightthickness': 1,
            'wrap': 'none',
        }
        if width is not None:
            text_area_config.update({
                'width': (width - 20) >> 1,
            })
        
        if height is not None:
            text_area_config.update({
                'height': height,
            })
        
        self.text_in_area = TtkbScrolledText(self, **text_area_config)

        if 'Darwin' == OS_TYPE:
            self.control_key_name = 'Cmd'
        else:
            self.control_key_name = 'Ctrl'

        self.buttons_frame = ttkb.Frame(self)
        self.run_button_text = '>>'
        self.run_button_width = len(self.run_button_text) + 1
        self.run_button = ttkb.Button(self.buttons_frame, text=self.run_button_text, width=self.run_button_width, command=self.run)
        self.run_button_tooltip = ToolTipHovered(self.run_button, f'<{self.control_key_name}+Enter> Run and clear input field. Prevents human missclicks')

        self.run_button_2_text = 'R>'
        self.run_button_2_width = len(self.run_button_2_text) + 1
        self.run_button_2 = ttkb.Button(self.buttons_frame, text=self.run_button_2_text, width=self.run_button_2_width, command=self.run_2)
        self.run_button_2_tooltip = ToolTipHovered(self.run_button_2, f'<{self.control_key_name}+Alt+Enter> Run')

        self.stop_button_text = 'X'
        self.stop_button_width = len(self.stop_button_text) + 1
        self.stop_button = ttkb.Button(self.buttons_frame, text=self.stop_button_text, width=self.stop_button_width, command=self.kill)
        self.stop_button['state'] = 'disabled'
        self.stop_button_tooltip = ToolTipHovered(self.stop_button, f'<{self.control_key_name}+x> Stop execution')

        self.text_out_area = ReadOnlyText(self, **text_area_config)


        self.func_header_read_only = ttkb.Label(self, text=coro_holder_header, font=('Helvetica', 8, 'bold'))
        self.func_header_read_only.pack(fill='x', expand='yes', side='top')

        self.text_in_area.pack(fill='both', side='left')
        self.run_button.pack(fill='both', side='top', pady=1)
        self.run_button_2.pack(fill='both', side='top', pady=1)
        self.stop_button.pack(fill='both', side='top', pady=1)
        self.buttons_frame.pack(fill='both', side='left')
        self.text_out_area.pack(fill='both', expand='yes', side='left')

        self.text_in_area._text.bind("<Control-Return>", self.ctrl_enter)
        self.text_in_area._text.bind("<Control-Alt-Return>", self.ctrl_alt_enter)
        self.text_in_area._text.bind("<Control-x>", self.ctrl_x)

    def ctrl_enter(self, event):
        self.run()
        return 'break'

    def ctrl_alt_enter(self, event):
        self.run_2()
        return 'break'

    def ctrl_x(self, event):
        self.kill()
        return 'break'

    def run(self, clear_input: bool = True):
        text = self.text_in_area._text.get('1.0', 'end')
        if clear_input:
            self.text_in_area._text.delete('1.0', 'end')
        
        self.text_out_area._text.insert('end', text)
        
        # disable the intput
        self.run_button['text'] = '...'
        self.run_button_width = len(self.run_button['text']) + 1
        self.run_button['state'] = 'disabled'
        self.run_button_2['text'] = '...'
        self.run_button_2_width = len(self.run_button_2['text']) + 1
        self.run_button_2['state'] = 'disabled'
        self.stop_button['state'] = 'normal'
        self.text_in_area._text['state'] = 'disabled'
        self.text_in_area.state(["disabled"])
        
        self.current_coro_id = current_interface()(PutCoro, self.eval_text, text)
    
    def run_2(self):
        return self.run(clear_input = False)

    def kill(self):
        if self.current_coro_id is not None:
            graceful_coro_destroyer(current_interface(), None, self.current_coro_id, first_phase_is_wait=False)
            self.current_coro_id = None
            self.run_button['text'] = self.run_button_text
            self.run_button_width = self.run_button_width
            self.run_button['state'] = 'normal'
            self.run_button_2['text'] = self.run_button_2_text
            self.run_button_2_width = self.run_button_2_width
            self.run_button_2['state'] = 'normal'
            self.stop_button['state'] = 'disabled'
            self.text_in_area._text['state'] = 'normal'
            self.text_in_area.state(["!disabled"])
    
    def eval_text(self, i: Interface, text):
        self.text_out_area._text.delete('1.0', 'end')

        # evalueate the text
        try:
            result = None
            local_vars = {
                'i': i,
                'result': result,
                'aggregator_key': command_executor_aggregator_view_key,
                'aggregator_append_key': command_executor_aggregator_append_view_key,
            }
            text = coro_holder_template.format(coro_body=prepare_coro_body(text))
            exec(text, local_vars)
            self.text_out_area._text.insert('1.0', f'{local_vars["result"]}')
        except GracefulCoroDestroy as ex:
            # self.text_out_area._text.insert('1.0', f'{type(ex)}')
            pass
        except GreenletExit as ex:
            # self.text_out_area._text.insert('1.0', f'{type(ex)}')
            pass
        except:
            self.text_out_area._text.insert('1.0', f'{exception_to_printable_text(get_exception())}')
        finally:
            # enable the intput
            self.run_button['text'] = self.run_button_text
            self.run_button_width = len(self.run_button['text']) + 1
            self.run_button['state'] = 'normal'
            self.run_button_2['text'] = self.run_button_2_text
            self.run_button_2_width = len(self.run_button_2['text']) + 1
            self.run_button_2['state'] = 'normal'
            self.stop_button['state'] = 'disabled'
            self.text_in_area._text['state'] = 'normal'
            self.text_in_area.state(["!disabled"])


class CSStatsFormatterMultiprocess:
    def __init__(self, filtered_paths: List[Sequence[str]] = None):
        self.filtered_paths: List[Sequence[str]] = [
            ['CoroSchedulerBase', 'loop', 'coroutines execution times'],
            ['CoroSchedulerBase', 'loop', 'longest continuous execution time of coroutines'],
            ['CoroSchedulerGreenlet', 'loop', 'coroutines execution times'],
            ['CoroSchedulerGreenlet', 'loop', 'longest continuous execution time of coroutines'],
            ['CoroSchedulerAwaitable', 'loop', 'coroutines execution times'],
            ['CoroSchedulerAwaitable', 'loop', 'longest continuous execution time of coroutines'],
        ] if filtered_paths is None else filtered_paths
        self._stop: bool = False
        self._subprocess_started: bool = False
        settings = SubprocessWorkerSettings()
        settings.initiation_function = self.process_initializer
        settings.working_function = self.process_worker
        settings.transport = Transport.queue
        settings.sendable_data_type = SendableDataType.marshalable
        settings.use_internal_subprocess_input_buffer = True
        settings.initialization_data = {
            'filtered_paths': self.filtered_paths
        }
        
        self.worker_lock: Lock = Lock(f'CSStatsFormatterMultiprocess.worker__{uuid4()}')
        self.worker: SubprocessWorker = SubprocessWorker(settings)
        self.worker.start(wait_for_process_readyness=False)
        self.worker_is_ready: bool = False
    
    async def _worker_readyness_waiting_coro(self, i: Interface):
        need_to_wait = True
        while need_to_wait:
            try:
                async with self.worker_lock:
                    self.worker.wait_for_subprocess_readines(block=False)

                need_to_wait = False
            except SubprocessIsNotReadyError:
                await i(Sleep, 0.1)
        
        self.worker_is_ready = True
    
    async def wait_for_worker_readyness(self, i: Interface):
        while not self.worker_is_ready:
            await i(Sleep, 0.1)
    
    def start(self, wr: Optional[TkObjWrapper] = None):
        if wr is None:
            i = current_interface()
            i(PutCoro, self._worker_readyness_waiting_coro)
        else:
            wr.put_coro(self._worker_readyness_waiting_coro)
    
    def stop(self):
        self._stop = True
        need_to_block = False
        try:
            i: Interface = current_interface()
        except:
            need_to_block = True
        
        with self.worker_lock:
            if need_to_block:
                self.worker.stop()
            else:
                run_in_thread_pool_fast(i, self.worker.stop)
        
    def update_filtered_paths(self, filtered_paths: List[Sequence[str]]):
        i: Interface = current_interface()
        self.filtered_paths = filtered_paths
        with self.worker_lock:
            if not self._stop:
                self.worker.send_data_to_subprocess({
                    'type': 'filter',
                    'data': self.filtered_paths
                })
    
    def __call__(self, data):
        with self.worker_lock:
            if not self.worker_is_ready:
                # return data
                return '<<Waiting for subprocess to start...>>'
            
            i: Interface = current_interface()
            self.worker.send_data_to_subprocess({
                'type': 'data',
                'data': data
            })
            result = None
            got_result = False
            while (not got_result) and (not self._stop):
                try:
                    result = self.worker.get_answer_from_subprocess(block=False)
                except Empty:
                    pass
                else:
                    got_result = True
                
                if got_result:
                    if 'data' != result['type']:
                        got_result = False
                
                if not got_result:
                    i(Sleep, 0.01)
            
            result = '' if result is None else result['data']
            result = f'{i.coro_id}. CoroScheduler stats:\n{result}'
            return result
    
    @staticmethod
    def process_initializer(init_data) -> Any:
        from pprintpp import pformat as pf
        return init_data
    
    @staticmethod
    def process_worker(global_data, data_msg):
        data = data_msg['data']
        if 'filter' == data_msg['type']:
            if global_data is None:
                global_data = dict()
            
            global_data['filtered_paths'] = data
            return None  # no answer to parent process
        elif 'data' == data_msg['type']:
            filtered_paths = global_data['filtered_paths']
            for path in filtered_paths:
                try_del_dict_item(data, path)
            
            data = pf(data, indent=4, width=120)
            data_msg['data'] = data
            return data_msg


class CorosMaxExecutionTimesProvider:
    def __init__(self, lifetime_stats_key: Hashable, stats_key: Hashable, period: float):
        self.lifetime_stats_key: Hashable = lifetime_stats_key
        self.stats_key: Hashable = stats_key
        self.period: float = period
        self.i: Interface = None
        self.fac = FastAggregatorClient()
        self._stop = False
        self.lifetime_stats: Dict[CoroID, float] = dict()
    
    def start(self):
        if self.i is None:
            self.i = current_interface()
        
        self.i(PutCoro, self._update)
    
    def stop(self):
        self._stop = True
    
    async def _update(self, i: Interface):
        while not self._stop:
            data = copy(i._loop.coro_longest_execution_time)
            new_coro_longest_execution_time = dict()
            for key, value in data.items():
                new_coro_longest_execution_time[key] = 0.0
                
                if key not in self.lifetime_stats:
                    self.lifetime_stats[key] = 0.0
                    
                self.lifetime_stats[key] = max(self.lifetime_stats[key], value)
            
            i._loop.coro_longest_execution_time = new_coro_longest_execution_time
            self.fac(self.lifetime_stats_key, self.lifetime_stats)
            self.fac(self.stats_key, data)
            await i(Sleep, self.period)
    
    def _formatter(self, title: str, data):
        i: Interface = current_interface()
        return f'{i.coro_id}. {title}:\n' + '\n'.join([f'{coro_id}: {value * 1000}' for coro_id, value in data.items()])
    
    def stats_formatter(self, data):
        return self._formatter('Coros Max Execution Times', data)
    
    def lifetime_stats_formatter(self, data):
        return self._formatter('Coros Lifetime Max Execution Times', data)


class CoroLogsAppendFormatterParts(IntFlag):
    none = 0
    prompt_string = auto()
    time = auto()
    caller_info = auto()
    logging_level = auto()
    log = auto()
    coros_traceback = auto()
    file_name_and_line_number = auto()
    traceback = auto()
    all = prompt_string | time | caller_info | logging_level | log | coros_traceback | file_name_and_line_number | traceback


class CoroLogsAppendFormatter(AggregatorAppendFormatter):
    def __init__(self, initial_text: str, desired_parts: CoroLogsAppendFormatterParts = CoroLogsAppendFormatterParts.all) -> None:
        super().__init__(initial_text)
        self._prompt_string: str = f'>>> {"-"*80}'
        self.desired_parts: CoroLogsAppendFormatterParts = desired_parts

    def __call__(self, data: List[Tuple[Tuple, Dict, Dict[str, Any]]]) -> Any:
        if data:
            data_part = '\n\n'.join([self._format(*log_info) for log_info in data])
            return super().__call__(f'\n{data_part}\n')
        else:
            return super().__call__(str())

    def _format(self, args, kwargs, info=None) -> str:
        if info is None:
            return f'{self._prompt_string}\n{args_kwargs_to_str(args, kwargs)}'
        else:
            output_strings: List[str] = list()
            if CoroLogsAppendFormatterParts.prompt_string in self.desired_parts:
                output_strings.append(self._prompt_string)

            if CoroLogsAppendFormatterParts.time in self.desired_parts:
                output_strings.append(f'> Time: {info[InfoFields.current_time]}; Perf Counter: {info[InfoFields.perf_counter_time]:17.6f}')

            if CoroLogsAppendFormatterParts.caller_info in self.desired_parts:
                output_strings.append(f'> λ: {info[InfoFields.caller_info]}')

            if CoroLogsAppendFormatterParts.logging_level in self.desired_parts:
                if InfoFields.logging_level in info:
                    output_strings.append(f'> Logging level: {info[InfoFields.logging_level]}')
            
            if CoroLogsAppendFormatterParts.log in self.desired_parts:
                output_strings.append(args_kwargs_to_str(args, kwargs))

            if CoroLogsAppendFormatterParts.coros_traceback in self.desired_parts:
                coro_parents_strings: List[str] = info[InfoFields.coro_parents_strings]
                if coro_parents_strings:
                    coro_parents_text: str = '\n'.join(coro_parents_strings)
                    output_strings.append(f'> Coros traceback:\n{coro_parents_text}')
            
            if CoroLogsAppendFormatterParts.file_name_and_line_number in self.desired_parts:
                output_strings.append(f'> @ {info[InfoFields.file_name]}:{info[InfoFields.line_number]}')

            if CoroLogsAppendFormatterParts.traceback in self.desired_parts:
                traceback_strings: List[str] = info[InfoFields.traceback_strings]
                if traceback_strings:
                    traceback_text: str = '\n'.join(traceback_strings)
                    traceback_text = traceback_text.strip('\n')
                    output_strings.append(f'> Traceback:\n{traceback_text}')
            
            return '\n'.join(output_strings)


class CorosLogsProvider:
    def __init__(self, coros_logs_key: Hashable, logs_limit: Union[None, int], period: float, force_stop_timeout: RationalNumber = 0.1):
        self.coros_logs_key: Hashable = coros_logs_key
        self.logs_limit: Union[None, int] = logs_limit
        self.period: float = period
        self.i: Interface = None
        self._update_coro_id: CoroID = None
        self.force_stop_timeout: RationalNumber = force_stop_timeout
        self.fac = FastAggregatorClient()
        self._stop = False
        self.lifetime_stats: Dict[CoroID, float] = dict()
        self._current_logs_taken: bool = False
        self._unsend_data: List[List[Tuple[Tuple, Dict]]] = list()
        self._log_service_handler_was_added: bool = False
    
    def start(self):
        if self.i is None:
            self.i = current_interface()
        
        self._update_coro_id = self.i(PutCoro, self._update)
    
    def stop(self):
        self._stop = True
        current_interface()(PutCoro, self._force_stop)
    
    async def _force_stop(self, i: Interface):
        if self._update_coro_id is not None:
            update_coro_id = self._update_coro_id
            self._update_coro_id = None
            await agraceful_coro_destroyer(i, self.force_stop_timeout, update_coro_id)
        
        if self._log_service_handler_was_added:
            self._log_service_handler_was_added = False
            await i(LogRequest().remove_iteration_handler(self))
        
        self._unsend_data = type(self._unsend_data)()
    
    async def _update(self, i: Interface):
        while not self._stop:
            if not self._log_service_handler_was_added:
                if i._loop.is_service_registered(Log):
                    self._log_service_handler_was_added = True
                    await i(LogRequest().add_iteration_handler(self))
                    
            logs_parts = self._unsend_data
            self._unsend_data = type(self._unsend_data)()
            for logs_part in logs_parts:
                self.fac(self.coros_logs_key, logs_part)
            
            await i(Sleep, self.period)
    
    def __call__(self, log_service: Log, data: List[Tuple[Tuple, Dict, Dict[str, Any]]], current_time: float, current_time_str: str):
        if self._current_logs_taken:
            if self.logs_limit is None:
                if data:
                    self._unsend_data.append(data)
            else:
                num_of_an_initial_logs_needed: int = self.logs_limit - len(data)
                if 0 <= num_of_an_initial_logs_needed:
                    if data:
                        self._unsend_data.append(data)
                else:
                    part_of_data = data[-self.logs_limit:]
                    if part_of_data:
                        self._unsend_data.append(part_of_data)
        else:
            self._current_logs_taken = True
            if self.logs_limit is None:
                combined_data = log_service.get_last_n_logs(None) + data
                if combined_data:
                    self._unsend_data.append(combined_data)
            else:
                num_of_an_initial_logs_needed: int = self.logs_limit - len(data)
                if 0 < num_of_an_initial_logs_needed:
                    combined_data = log_service.get_last_n_logs(num_of_an_initial_logs_needed) + data
                    if combined_data:
                        self._unsend_data.append(combined_data)
                if 0 == num_of_an_initial_logs_needed:
                    if data:
                        self._unsend_data.append(data)
                else:
                    part_of_data = data[-self.logs_limit:]
                    if part_of_data:
                        self._unsend_data.append(part_of_data)


class SchedulerPerformanceFormatter:
    def __init__(self, loop_iteration_delta_times_key: Hashable, loop_iteration_delta_times_lifetime_stats_key: Hashable, loop_iteration_delta_times_stats_key: Hashable, window_size: int):
        self.external_items_key: Hashable = loop_iteration_delta_times_key
        self.internal_items_key: Hashable = f'internal_scheduler_tdelta__{uuid4()}'
        self.lifetime_stats_key: Hashable = loop_iteration_delta_times_lifetime_stats_key
        self.stats_key: Hashable = loop_iteration_delta_times_stats_key
        self.window_size: int = window_size
        self.fac = FastAggregatorClient()
        self.i: Interface = None
        self.data_window: List = list()
        self.max_deviation: float = None
        self.min_deviation: float = None
        self.max_iter_per_sec: float = None
        self.min_iter_per_sec: float = None
        self._stop: bool = False

        settings = SubprocessWorkerSettings()
        settings.initiation_function = self.process_initializer
        settings.working_function = self.process_worker
        settings.transport = Transport.queue
        settings.sendable_data_type = SendableDataType.marshalable
        self.worker_lock: Lock = Lock(f'SchedulerPerformanceFormatter.worker__{uuid4()}')
        self.worker: SubprocessWorker = SubprocessWorker(settings)
        self.worker.start(wait_for_process_readyness=False)
        self.worker_is_ready: bool = False
    
    async def _worker_readyness_waiting_coro(self, i: Interface):
        need_to_wait = True
        while need_to_wait:
            try:
                async with self.worker_lock:
                    self.worker.wait_for_subprocess_readines(block=False)
                
                need_to_wait = False
            except SubprocessIsNotReadyError:
                await i(Sleep, 0.1)
        
        self.worker_is_ready = True
    
    async def wait_for_worker_readyness(self, i: Interface):
        while not self.worker_is_ready:
            await i(Sleep, 0.1)
    
    def start(self, wr: Optional[TkObjWrapper] = None):
        if self.i is None:
            self.i = current_interface()
        
        if wr is None:
            self.i(PutCoro, self._worker_readyness_waiting_coro)
            self.i(PutCoro, self._update)
            self.i(PutCoro, self._update_stats)
        else:
            wr.put_coro(self._worker_readyness_waiting_coro)
            wr.put_coro(self._update)
            wr.put_coro(self._update_stats)
    
    def stop(self):
        self._stop = True
        need_to_block = False
        try:
            i: Interface = current_interface()
        except:
            need_to_block = True
        
        with self.worker_lock:
            if need_to_block:
                self.worker.stop()
            else:
                run_in_thread_pool_fast(i, self.worker.stop)
    
    def _update(self, i: Interface):
        i(RunCoro, self.wait_for_worker_readyness)
        while not self._stop:
            # start = perf_counter()
            i(Yield)
            # stop = perf_counter()
            # delta_time = stop - start
            loop: CoroSchedulerType = i._loop
            delta_time = loop.loop_iteration_delta_time
            self.fac(self.external_items_key, delta_time)
            self.fac(self.internal_items_key, delta_time)

    def _update_stats(self, i: Interface):
        i(RunCoro, self.wait_for_worker_readyness)
        while not self._stop:
            try:
                data = i(FastAggregator, self.internal_items_key)
                if len(data) >= self.window_size:
                    self.data_window = data[-self.window_size:]
                else:
                    additional_data_len = self.window_size - len(data)
                    if len(self.data_window) >= additional_data_len:
                        self.data_window = self.data_window[-additional_data_len:] + data
                    else:
                        self.data_window.extend(data)
                
                iter_per_sec, val_99, val_95, val_68, max_deviation, min_deviation = self.compute_stats(i, self.data_window)
                if self.max_deviation is None:
                    self.max_deviation = max_deviation
                else:
                    self.max_deviation = max(self.max_deviation, max_deviation)
                
                if self.min_deviation is None:
                    self.min_deviation = min_deviation
                else:
                    self.min_deviation = min(self.min_deviation, min_deviation)
                
                if self.max_iter_per_sec is None:
                    self.max_iter_per_sec = iter_per_sec
                else:
                    self.max_iter_per_sec = max(self.max_iter_per_sec, iter_per_sec)
                
                if self.min_iter_per_sec is None:
                    self.min_iter_per_sec = iter_per_sec
                else:
                    self.min_iter_per_sec = min(self.min_iter_per_sec, iter_per_sec)
                
                self.fac(self.stats_key, (iter_per_sec, val_99, val_95, val_68, max_deviation, min_deviation))
                self.fac(self.lifetime_stats_key, (self.max_iter_per_sec, self.min_iter_per_sec, self.max_deviation, self.min_deviation))
                
                self.window_size = 2 * int(round(iter_per_sec))
            except KeyError:
                pass
            
            i(Sleep, 0.2)
    
    def compute_stats(self, i: Interface, data):
        with self.worker_lock:
            self.worker.send_data_to_subprocess(data)
            result = None
            got_result = False
            while not got_result:
                try:
                    result = self.worker.get_answer_from_subprocess(block=False)
                except Empty:
                    pass
                else:
                    got_result = True
                
                if not got_result:
                    i(Sleep, 0.005)
                
            return result
    
    @staticmethod
    def process_initializer(init_data):
        pass
    
    @staticmethod
    def process_worker(global_data, data):
        from cengal.statistics.normal_distribution import count_99_95_68
        val_99, val_95, val_68, max_deviation, min_deviation = count_99_95_68(data)
        average_data = average(data)
        iter_per_sec = 0 if 0 == average_data else (1 / average_data)
        return (iter_per_sec, val_99, val_95, val_68, max_deviation, min_deviation)
        
    @staticmethod
    def item_formatter(data) -> str:
        return f'{data}'
        
    @staticmethod
    def stats_formatter(data) -> str:
        i: Interface = current_interface()
        iter_per_sec, val_99, val_95, val_68, max_deviation, min_deviation = data
        return f'{i.coro_id}. Stats:\niter_per_sec: {iter_per_sec}\nval_99: {val_99}\nval_95: {val_95}\nval_68: {val_68}\nmax_deviation: {max_deviation}\nmin_deviation: {min_deviation}'
        
    @staticmethod
    def lifetime_stats_formatter(data) -> str:
        i: Interface = current_interface()
        max_iter_per_sec, min_iter_per_sec, max_deviation, min_deviation = data
        return f'{i.coro_id}. Lifetime Stats:\nmax_iter_per_sec: {max_iter_per_sec}\nmin_iter_per_sec: {min_iter_per_sec}\nmax_deviation: {max_deviation}\nmin_deviation: {min_deviation}'


class HelpDialog(simpledialog.Dialog):
    def body(self, master):
        self.title('Console Description and Help')

        text_area_config = {
            'highlightcolor': Style.instance.colors.primary,
            'highlightbackground': Style.instance.colors.border,
            'highlightthickness': 1,
            'wrap': 'none',
            'width': 120,
            'height': 30,
        }

        self.text_area = ReadOnlyText(self, **text_area_config)
        self.text_area.pack(fill='both', expand='yes')
        self.text_area._text.delete('1.0', 'end')
        self.text_area._text.insert('end', coro_holder_template)
        
        return self.text_area  # initial focus


class FilterSetupDialog(ttkb.Toplevel):
    def __init__(self, parent, filtered_paths):
        # position is centered to parent
        parent_pos_left, parent_pos_top = parent.winfo_x(), parent.winfo_y()
        parent_size_x, parent_size_y = parent.winfo_width(), parent.winfo_height()
        own_size_x, own_size_y = 600, 400
        own_pos_x = parent_pos_left + (parent_size_x - own_size_x) // 2
        own_pos_y = parent_pos_top + (parent_size_y - own_size_y) // 2
        super().__init__(parent, size=(own_size_x, own_size_y), resizable=(True, True), position=(own_pos_x, own_pos_y))
        self.filtered_paths = filtered_paths
        self.title('Filtered paths setup')
        self.transient(parent)
        self.result = None

        ttkb.Label(self, text="Enter your text:").pack(fill=None, expand='no', side='top')
        self.text_entry = TtkbScrolledText(self, width=30, height=10)
        self.text_entry.pack(fill='both', expand='yes', side='top')

        button_frame = ttkb.Frame(self)
        button_frame.pack(fill='x', expand='no', side='right')

        ttkb.Button(button_frame, text='OK', command=self.ok).pack(side=ttkb.LEFT, padx=1)
        ttkb.Button(button_frame, text='Cancel', command=self.cancel).pack(side=ttkb.LEFT, padx=1)

        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.text_entry.focus_set()
        filtered_paths_list = list()
        for path in self.filtered_paths:
            filtered_paths_list.append(dict_path_to_str(path))
        
        filtered_paths_text = '\n'.join(filtered_paths_list) + '\n'
        self.text_entry._text.insert('1.0', filtered_paths_text)

        self.text_entry._text.bind("<Control-Return>", self.ctrl_enter)
        self.text_entry._text.bind("<Escape>", self.escape)
        self.text_entry._text.focus_set()

    def ctrl_enter(self, event):
        self.ok(event)
        return 'break'

    def escape(self, event):
        self.cancel(event)
        return 'break'

    def ok(self, event=None):
        filtered_paths_text = self.text_entry.get('1.0', 'end-1c')
        result_lines = filtered_paths_text.split('\n')
        filtered_paths = list()
        for line in result_lines:
            if line:
                filtered_paths.append(srt_to_dict_path(line))
        
        self.result = filtered_paths
        self.destroy()

    def cancel(self, event=None):
        self.result = self.filtered_paths
        self.destroy()


# class Application(Tk):
class Application(ttkb.Window):
    def __init__(self, style: str = 'superhero', filtered_paths: List[List[str]] = None, current_children_pack_type: int = 1):
        super().__init__(size=(1900, 900), resizable=(True, True))
        self.style_name = style
        Style(style)
        self.current_children_pack_type: int = current_children_pack_type
        self.max_children_pack_type: int = 2
        self.packed_widgets: List[tk.Widget] = list()
        self.filtered_paths: List[Sequence[str]] = filtered_paths

        self.title('CoroScheduler Admin')
        self.resizable(width=True, height=True)

        self.sfmp = CSStatsFormatterMultiprocess(self.filtered_paths)
        
        self.scheduler_stats_frame = ttkb.Frame(self)
        self.scheduler_stats_control_frame = ttkb.Frame(self.scheduler_stats_frame)
        
        self.scheduler_stats_layout_button_text = self.scheduler_stats_layout_button_name()
        self.scheduler_stats_layout_button_text_len = len('L') + 1
        self.scheduler_stats_layout_button = ttkb.Button(self.scheduler_stats_control_frame, text=self.scheduler_stats_layout_button_text, width=self.scheduler_stats_layout_button_text_len, command=self.scheduler_stats_layout_button_on_click)
        self.scheduler_stats_layout_button_tooltip = ToolTipHovered(self.scheduler_stats_layout_button, 'Layout circle switch')

        self.scheduler_stats_format_button_text = 'F'
        self.scheduler_stats_format_button_text_len = len(self.scheduler_stats_format_button_text) + 1
        self.scheduler_stats_format_button = ttkb.Button(self.scheduler_stats_control_frame, text=self.scheduler_stats_format_button_text, width=self.scheduler_stats_format_button_text_len, command=self.scheduler_stats_format_button_on_click)
        self.scheduler_stats_format_button_tooltip = ToolTipHovered(self.scheduler_stats_format_button, 'Filter our unnecessary or flickering paths')
        
        self.scheduler_stats_help_button_text = '?'
        self.scheduler_stats_help_button_text_len = len(self.scheduler_stats_help_button_text) + 1
        self.scheduler_stats_help_button = ttkb.Button(self.scheduler_stats_control_frame, text=self.scheduler_stats_help_button_text, width=self.scheduler_stats_help_button_text_len, bootstyle="info", command=self.scheduler_stats_help_button_on_click)
        self.scheduler_stats_help_button_tooltip = ToolTipHovered(self.scheduler_stats_help_button, 'Console description and help')

        self.coro_scheduler_view = AggregatorView(False, False, 'coro_scheduler_stats', 1, self.sfmp, 25, 23, self.scheduler_stats_frame)

        self.coro_logs_formatter: CoroLogsAppendFormatter = CoroLogsAppendFormatter('Coro Logs')
        self.coro_logs_provider: CorosLogsProvider = CorosLogsProvider('coro scheduler logs', None, 0.25)
        self.coro_scheduler_logs_view = AggregatorView(True, True, 'coro scheduler logs', 1, self.coro_logs_formatter, 25, 23, self.scheduler_stats_frame)

        self.command_executor_view = CommandExecutor(160, 6)
        
        self.cmet = CorosMaxExecutionTimesProvider('coros lifetime max execution times', 'coros max execution times', 0.5)
        
        self.coros_lifetime_max_execution_times = AggregatorView(False, False, 'coros lifetime max execution times', 0.75, self.cmet.lifetime_stats_formatter, 25, 13, self)
        
        self.coros_max_execution_times = AggregatorView(False, False, 'coros max execution times', 0.75, self.cmet.stats_formatter, 25, 13, self)
        
        self.spf = SchedulerPerformanceFormatter('loop_iteration_delta_times', 'loop_iteration_delta_times_lifetime_stats', 'loop_iteration_delta_times_stats', 5000)
        
        self.scheduler_tdelta_formatter = AggregatorAppendFormatter('Scheduler TDelta')
        self.scheduler_tdelta = AggregatorView(True, False, 'loop_iteration_delta_times', 0.1, self.scheduler_tdelta_formatter, 22, 13, self)
        self.scheduler_tdelta.default_auto_scroll = False
        self.scheduler_tdelta.max_len = 1000
        
        def aggregator_view_formatter(data: Any) -> str:
            return f'{current_interface().coro_id}. Aggregator:\n{data}'
        
        self.command_executor_aggregator_view = AggregatorView(False, False, command_executor_aggregator_view_key, 1, aggregator_view_formatter, 25, 13, self)
        FastAggregatorClient()(command_executor_aggregator_view_key, str())
        
        self.command_executor_aggregator_append_formatter = AggregatorAppendFormatter('Aggregator Append')
        self.command_executor_aggregator_append_view = AggregatorView(True, True, command_executor_aggregator_append_view_key, 1, self.command_executor_aggregator_append_formatter, 25, 13, self)
        self.command_executor_aggregator_append_view.max_len = 5000
        FastAggregatorClient()(command_executor_aggregator_append_view_key, str())

        self.scheduler_lifetime_stats = AggregatorView(False, False, 'loop_iteration_delta_times_lifetime_stats', 0.3, self.spf.lifetime_stats_formatter, 36, 5, self)
        
        self.scheduler_stats = AggregatorView(False, False, 'loop_iteration_delta_times_stats', 0.3, self.spf.stats_formatter, 36, 7, self)

        self.pack_children(self.current_children_pack_type)

        # set window size taking into account the size of the widgets
        self.update()
        width = self.winfo_width()
        height = self.winfo_height()
        x = (self.winfo_screenwidth() // 2) - (width // 2)
        y = (self.winfo_screenheight() // 2) - (height // 2)
        self.geometry('{}x{}+{}+{}'.format(width, height, x, y))
        self.update()
    
    def pack_forget_children(self):
        self.packed_widgets.reverse()
        for widget in self.packed_widgets:
            widget.pack_forget()
    
    def pack_children(self, pack_type: int):
        self.pack_forget_children()
        self.packed_widgets.clear()
        if 0 == pack_type:
            self.scheduler_stats_layout_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_layout_button)
            self.scheduler_stats_format_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_format_button)
            self.scheduler_stats_help_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_help_button)
            self.scheduler_stats_control_frame.pack(fill='y', expand='no', side='left')
            self.packed_widgets.append(self.scheduler_stats_control_frame)
            self.coro_scheduler_view.pack(fill='both', expand='yes', side='left')
            self.packed_widgets.append(self.coro_scheduler_view)
            self.coro_scheduler_logs_view.pack(fill='both', expand='no', side='left')
            self.packed_widgets.append(self.coro_scheduler_logs_view)

            self.scheduler_stats_frame.pack(fill='both', expand='yes', side='bottom')
            self.packed_widgets.append(self.scheduler_stats_frame)
            self.command_executor_view.pack(fill='x', expand='no', side='bottom')
            self.packed_widgets.append(self.command_executor_view)
            self.coros_lifetime_max_execution_times.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.coros_lifetime_max_execution_times)
            self.coros_max_execution_times.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.coros_max_execution_times)
            self.scheduler_tdelta.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.scheduler_tdelta)
            self.command_executor_aggregator_view.pack(fill='x', expand='yes', side='right')
            self.packed_widgets.append(self.command_executor_aggregator_view)
            self.command_executor_aggregator_append_view.pack(fill='x', expand='yes', side='right')
            self.packed_widgets.append(self.command_executor_aggregator_append_view)
            self.scheduler_lifetime_stats.pack(fill=None, expand='no', side='top')
            self.packed_widgets.append(self.scheduler_lifetime_stats)
            self.scheduler_stats.pack(fill=None, expand='no', side='bottom')
            self.packed_widgets.append(self.scheduler_stats)
        elif 1 == pack_type:
            self.scheduler_stats_layout_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_layout_button)
            self.scheduler_stats_format_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_format_button)
            self.scheduler_stats_help_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_help_button)
            self.scheduler_stats_control_frame.pack(fill='y', expand='no', side='left')
            self.packed_widgets.append(self.scheduler_stats_control_frame)
            self.coro_scheduler_view.pack(fill='both', expand='yes', side='left')
            self.packed_widgets.append(self.coro_scheduler_view)
            self.coro_scheduler_logs_view.pack(fill='both', expand='yes', side='left')
            self.packed_widgets.append(self.coro_scheduler_logs_view)

            self.scheduler_stats_frame.pack(fill='both', expand='yes', side='bottom')
            self.packed_widgets.append(self.scheduler_stats_frame)
            self.command_executor_view.pack(fill='x', expand='no', side='bottom')
            self.packed_widgets.append(self.command_executor_view)
            self.coros_lifetime_max_execution_times.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.coros_lifetime_max_execution_times)
            self.coros_max_execution_times.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.coros_max_execution_times)
            self.scheduler_tdelta.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.scheduler_tdelta)
            self.command_executor_aggregator_view.pack(fill='x', expand='yes', side='right')
            self.packed_widgets.append(self.command_executor_aggregator_view)
            self.command_executor_aggregator_append_view.pack(fill='x', expand='yes', side='right')
            self.packed_widgets.append(self.command_executor_aggregator_append_view)
            self.scheduler_lifetime_stats.pack(fill=None, expand='no', side='top')
            self.packed_widgets.append(self.scheduler_lifetime_stats)
            self.scheduler_stats.pack(fill=None, expand='no', side='bottom')
            self.packed_widgets.append(self.scheduler_stats)
        elif 2 == pack_type:
            self.scheduler_stats_layout_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_layout_button)
            self.scheduler_stats_format_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_format_button)
            self.scheduler_stats_help_button.pack(fill=None, expand='no', side='top', pady=1)
            self.packed_widgets.append(self.scheduler_stats_help_button)
            self.scheduler_stats_control_frame.pack(fill='y', expand='no', side='left')
            self.packed_widgets.append(self.scheduler_stats_control_frame)
            self.coro_scheduler_view.pack(fill='both', expand='no', side='left')
            self.packed_widgets.append(self.coro_scheduler_view)
            self.coro_scheduler_logs_view.pack(fill='both', expand='yes', side='left')
            self.packed_widgets.append(self.coro_scheduler_logs_view)

            self.scheduler_stats_frame.pack(fill='both', expand='yes', side='bottom')
            self.packed_widgets.append(self.scheduler_stats_frame)
            self.command_executor_view.pack(fill='x', expand='no', side='bottom')
            self.packed_widgets.append(self.command_executor_view)
            self.coros_lifetime_max_execution_times.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.coros_lifetime_max_execution_times)
            self.coros_max_execution_times.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.coros_max_execution_times)
            self.scheduler_tdelta.pack(fill='x', expand='no', side='left')
            self.packed_widgets.append(self.scheduler_tdelta)
            self.command_executor_aggregator_view.pack(fill='x', expand='yes', side='right')
            self.packed_widgets.append(self.command_executor_aggregator_view)
            self.command_executor_aggregator_append_view.pack(fill='x', expand='yes', side='right')
            self.packed_widgets.append(self.command_executor_aggregator_append_view)
            self.scheduler_lifetime_stats.pack(fill=None, expand='no', side='top')
            self.packed_widgets.append(self.scheduler_lifetime_stats)
            self.scheduler_stats.pack(fill=None, expand='no', side='bottom')
            self.packed_widgets.append(self.scheduler_stats)
        else:
            raise NotImplementedError
    
    def scheduler_stats_layout_button_name(self) -> str:
        return 'L' if 0 == self.current_children_pack_type else 'L' + str(self.current_children_pack_type)

    def scheduler_stats_layout_button_on_click(self):
        self.current_children_pack_type += 1
        if self.current_children_pack_type > self.max_children_pack_type:
            self.current_children_pack_type = 0
        
        self.scheduler_stats_layout_button_text = self.scheduler_stats_layout_button_name()
        self.scheduler_stats_layout_button.config(text=self.scheduler_stats_layout_button_text)
        self.pack_children(self.current_children_pack_type)
    
    def scheduler_stats_format_button_on_click(self):
        d = FilterSetupDialog(self, self.sfmp.filtered_paths)
        self.wait_window(d)
        self.sfmp.update_filtered_paths(d.result)
    
    def scheduler_stats_help_button_on_click(self):
        HelpDialog(self)

    def start(self, wr: TkObjWrapper):
        self.spf.start()
        self.sfmp.start(wr)
        self.coro_logs_provider.start()
        self.scheduler_tdelta.start(wr)
        self.command_executor_aggregator_view.start(wr)
        self.command_executor_aggregator_append_view.start(wr)
        self.scheduler_lifetime_stats.start(wr)
        self.scheduler_stats.start(wr)
        self.coro_scheduler_view.start(wr)
        self.coro_scheduler_logs_view.start(wr)
        self.cmet.start()
        self.coros_lifetime_max_execution_times.start(wr)
        self.coros_max_execution_times.start(wr)
    
    def stop(self):
        self.spf.stop()
        self.sfmp.stop()
        self.coro_logs_provider.stop()
        self.scheduler_tdelta.stop()
        self.command_executor_aggregator_view.stop()
        self.command_executor_aggregator_append_view.stop()
        self.scheduler_lifetime_stats.stop()
        self.scheduler_stats.stop()
        self.coro_scheduler_view.stop()
        self.coro_scheduler_logs_view.stop()
        self.cmet.stop()
        self.coros_lifetime_max_execution_times.stop()
        self.coros_max_execution_times.stop()
    
    def prepare(self, i: Interface, wr: TkObjWrapper):
        i(Yield)
        self.start(wr)
        i(Yield)


def coro_scheduler_admin__view(i: Interface, on_close: Optional[AnyWorker] = None, app_args_kwargs: Optional[ArgsKwargs] = None):
    app_args, app_kwargs = app_args_kwargs() if app_args_kwargs is not None else ArgsKwargs()()
    with TkinterContextManager(i, Application(*app_args, **app_kwargs)) as wr:
        app: Application = cast(Application, wr.tk)
        app.prepare(i, wr)
        i(InstanceRequest().set('admin_tk_app', app))
    
    if on_close is not None:
        app.stop()
        i(RunCoro, on_close)


def scheduler_stats_aggregator_provider(i: Interface, fac_key: str = 'coro_scheduler_stats'):
    cs: CoroSchedulerType = i._loop
    fac = FastAggregatorClient()
    while True:
        stats_level: EntityStatsMixin.StatsLevel = EntityStatsMixin.StatsLevel.info
        result = dict()
        name, stats = cs.get_entity_stats(stats_level)
        result[name] = stats
        fac(fac_key, result)
        i(Sleep, 0.5)


def cs_init(cs: CoroSchedulerType):
    cs.set_coro_time_measurement(True)
    cs.set_coro_history_gathering(True)
    cs.set_loop_iteration_time_measurement(True)


async def start_admin(i: Interface, on_close: Optional[AnyWorker] = None, app_args_kwargs: Optional[ArgsKwargs] = None):
    cs_init(i._loop)
    await i(PutCoroRequest().turn_on_tree_monitoring(True))
    await i(PutCoro, scheduler_stats_aggregator_provider)
    await i(PutCoro, coro_scheduler_admin__view, on_close, app_args_kwargs)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class DirTemplatesManager:
    def __init__(self, templates):
        self._templates = templates

    def make_an_absolute_template(self, root, template_name):
        template_content = self._templates[template_name]
        absolute_template_content = dict()
        for key, data in template_content.items():
            data = data.split(os.path.sep)
            data = os.path.join(root, *data)
            absolute_template_content[key] = data
        return absolute_template_content

    def create_template_in_the_folder(self, root, template_name):
        absolute_template = self.make_an_absolute_template(root, template_name)
        for key, path in absolute_template:
            if not os.path.isdir(path):
                os.makedirs(path)


def get_property_as_relative_dir(config_manager, file, property_name, root):
    property_data = config_manager.get_property(file, property_name)
    property_data = property_data.split(os.path.sep)
    result = os.path.join(root, *property_data)
    return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import gc
from cengal.code_flow_control.gc import DisableGC
from contextlib import contextmanager
from cengal.code_flow_control.smart_values.versions.v_2 import ValueExistence
from cengal.parallel_execution.coroutines.coro_standard_services.lazy_print.versions.v_0.lazy_print import lprint
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.time_management.repeat_for_a_time import Tracer, ClockType
from typing import Generator, Any

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class PerformanceTestResult(Exception):
    def __init__(self, result):
        super(PerformanceTestResult, self).__init__()
        self.result = result


@contextmanager
def test_run_time(test_name: str, number_of_iterations: int, throw_result: bool=False, throw_result_anyway: bool=True, ignore_index=False) -> Generator[ValueExistence[int], Any, Any]:
    index = ValueExistence(True, copy.copy(number_of_iterations))
    start_time = perf_counter()
    exception_occures = False
    try:
        yield index
    except:
        if not throw_result_anyway:
            exception_occures = True
        raise
    finally:
        if not ignore_index:
            number_of_iterations -= index.value
        
        end_time = perf_counter()
        result_time = end_time - start_time
        if result_time > 0:
            text_result = f'>>> "{test_name}"\n\tIt was used {result_time} seconds to process {number_of_iterations} iterations.\n\tThere is {number_of_iterations / result_time} iterations per second\n'
        else:
            text_result = f'>>> "{test_name}"\n\tIt was used {result_time} seconds to process {number_of_iterations} iterations.\n'

        lprint(text_result)

        if (not exception_occures) and throw_result:
            result_data = dict()
            result_data['test_name'] = test_name
            result_data['result_time'] = result_time
            if result_time > 0:
                result_data['iterations_per_time_unit'] = number_of_iterations / result_time
            else:
                result_data['iterations_per_time_unit'] = None
            raise PerformanceTestResult(result_data)


def test_function_run_time(testable_function):
    """
    Use 'performance_test_lib__iterations_qnt=1000000' parameter to pass number of iterations
    :param testable_function: function
    :return:
    """
    def run_time_test_func(*args, **kwargs):

        test_name = ''
        if 'performance_test_lib__test_name' in kwargs:
            test_name = str(kwargs['performance_test_lib__test_name'])
            del kwargs['performance_test_lib__test_name']
        test_name = '{}: {}'.format(str(testable_function), test_name)

        number_of_iterations = 1
        if 'performance_test_lib__iterations_qnt' in kwargs:
            number_of_iterations = int(kwargs['performance_test_lib__iterations_qnt'])
            del kwargs['performance_test_lib__iterations_qnt']

        throw_result = False
        if 'performance_test_lib__throw_result' in kwargs:
            throw_result = kwargs['performance_test_lib__throw_result']
            del kwargs['performance_test_lib__throw_result']

        with test_run_time(test_name, number_of_iterations, throw_result) as index:
            while index.value > 0:
                testable_function(*args, **kwargs)
                index.value -= 1
    return run_time_test_func


def process_performance_test_results(tracer: Tracer, test_name: str, throw_result: bool=False):
    number_of_iterations = tracer.iterations_made
    result_time = tracer.time_spent
    iterations_per_time_unit = tracer.iter_per_time_unit
    print('>>> "{}"'.format(test_name))
    print('\t' + 'It was used', result_time, 'seconds to process', number_of_iterations, 'iterations.')
    print('\t' + 'There is', iterations_per_time_unit, 'iterations per second')

    if throw_result:
        result_data = (test_name, result_time, iterations_per_time_unit)
        raise PerformanceTestResult(result_data)


@contextmanager
def test_performance(test_name: str, run_time: float, throw_result: bool=False, clock_type=ClockType.perf_counter):
    tracer = Tracer(run_time, clock_type)
    try:
        yield tracer
    except:
        raise
    finally:
        process_performance_test_results(tracer, test_name, throw_result)


def test_function_performance(testable_function):
    """
    Use 'performance_test_lib__run_time=1.5' parameter to pass number of seconds to test
    :param testable_function: function
    :return:
    """
    def run_time_test_func(*args, **kwargs):

        test_name = ''
        if 'performance_test_lib__test_name' in kwargs:
            test_name = str(kwargs['performance_test_lib__test_name'])
            del kwargs['performance_test_lib__test_name']
        test_name = '{}: {}'.format(str(testable_function), test_name)

        run_time = 0
        if 'performance_test_lib__run_time' in kwargs:
            run_time = int(kwargs['performance_test_lib__run_time'])
            del kwargs['performance_test_lib__run_time']

        throw_result = False
        if 'performance_test_lib__throw_result' in kwargs:
            throw_result = kwargs['performance_test_lib__throw_result']
            del kwargs['performance_test_lib__throw_result']

        clock_type = ClockType.perf_counter
        if 'performance_test_lib__clock_type' in kwargs:
            clock_type = kwargs['performance_test_lib__clock_type']
            del kwargs['performance_test_lib__clock_type']

        with test_performance(test_name, run_time, throw_result, clock_type) as tracer:
            while tracer.iter():
                testable_function(*args, **kwargs)

    return run_time_test_func


class PrecisePerformanceTestTracer(Tracer):
    """
    Precise tracer.
    At first you need to use it as a usual Tracer. After tracing was done - use it as a fast `for i in range(...)` block

    Example of use:

        tr = PrecisePerformanceTestTracer(10.0)
        while tr.iter():
            i = '456'
            k = int('1243' + i)

        with tr as fast_iter:
            for i in fast_iter:
                i = '456'
                k = int('1243' + i)

        print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    """

    def __init__(self,
                 run_time: float,
                 clock_type: ClockType=ClockType.perf_counter,
                 suppress_exceptions: bool=False,
                 turn_off_gc: bool=False
                 ):
        super().__init__(run_time, clock_type)
        self.suppress_exceptions = suppress_exceptions
        self.turn_off_gc = turn_off_gc
        self.gc_was_enabled = None

    def __enter__(self):
        self._relevant_start_time = self._start_time = self._relevant_stop_time = self._end_time = self._clock()
        self._relevant_number_of_iterations_at_start = 0
        if self.turn_off_gc:
            self.gc_was_enabled = gc.isenabled()
            gc.disable()
        
        return range(self._last_tracked_number_of_iterations)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._relevant_stop_time = self._end_time = self._clock()
        if self.turn_off_gc and self.gc_was_enabled:
            gc.enable()
        
        if self.suppress_exceptions:
            return True



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['Communication']

from cengal.parallel_execution.coroutines.coro_scheduler import *
from enum import Enum
from typing import Hashable, NoReturn, Dict, Tuple, Any


class Communication(Service):

    class Requests(Enum):
        send_async = 0        # Coro will get control immediately after message sent
        send_sync = 1         # Coro will get control only after response will be received
        read_async = 2        # Coro will get control immediately even if input queue is empty
        read_sync = 3         # Coro will get control only if input queue is not empty or when new response will be
        # received
        named_send_async = 4  # Coro will get control immediately after message sent
        named_send_sync = 5   # Coro will get control only after response will be received
        named_read_async = 6  # Coro will get control immediately even if input queue is empty
        named_read_sync = 7   # Coro will get control only if input queue is not empty or when new response will be
        # received

    class Request:
        def __init__(self):
            self.request_type = None  # type: Optional[int]
            self.args = None          # type: Optional[Tuple]
            self.kwargs = None        # type: Optional[Dict]

        def send_async(self, recipient_id: CoroID, message: Any) -> 'Communication.Request':
            self._save(0, recipient_id, message)
            return self

        def send_blocking(self, recipient_id: CoroID, message: Any) -> 'Communication.Request':
            self._save(1, recipient_id, message)
            return self

        def read_async(self) -> 'Communication.Request':
            self._save(2)
            return self

        def read_blocking(self) -> 'Communication.Request':
            self._save(3)
            return self

        def send_async_named(
                self, sender_id: Hashable, recipient_id: Hashable, message: Any) -> 'Communication.Request':
            self._save(4, sender_id, recipient_id, message)
            return self

        def send_blocking_named(
                self, sender_id: Hashable, recipient_id: Hashable, message: Any) -> 'Communication.Request':
            self._save(5, sender_id, recipient_id, message)
            return self

        def read_async_named(self, recipient_id: Hashable) -> 'Communication.Request':
            self._save(6, recipient_id)
            return self

        def read_blocking_named(self, recipient_id: Hashable) -> 'Communication.Request':
            self._save(7, recipient_id)
            return self

        def _save(self, __request__type__: int, *args, **kwargs):
            self.request_type = __request__type__
            self.args = args
            self.kwargs = kwargs

    def __init__(self, loop: CoroSchedulerType):
        super(Communication, self).__init__(loop)

    def single_task_registration_or_immediate_processing(
            self, request: 'Communication.Requests', *args, **kwargs
    ) -> Tuple[bool, Any]:
        if 0 == request.request_type:
            self.request_send_async(*args, **kwargs)
            return True, None, None
        elif 1 == request.request_type:
            self.request_send_sync(*args, **kwargs)
            return False, None, None
        elif 2 == request.request_type:
            result = self.request_read_async()
            return True, result, None
        elif 3 == request.request_type:
            self.request_read_sync()
            return False, None, None
        elif 4 == request.request_type:
            self.request_named_send_async(*args, **kwargs)
            return True, None, None
        elif 5 == request.request_type:
            self.request_named_send_sync(*args, **kwargs)
            return False, None, None
        elif 6 == request.request_type:
            result = self.request_named_read_async(*args, **kwargs)
            return True, result, None
        elif 7 == request.request_type:
            self.request_named_read_sync(*args, **kwargs)
            return False, None, None

    def request_send_async(self, recipient_id: CoroID, message: Any) -> NoReturn:
        raise NotImplementedError

    def request_send_sync(self, recipient_id: CoroID, message: Any) -> NoReturn:
        raise NotImplementedError

    def request_read_async(self) -> Tuple[CoroID, Any]:
        """
        Will return tuple with sender coro ID and message

        :rtype: Tuple[CoroID, Any]
        """
        raise NotImplementedError

    def request_read_sync(self) -> NoReturn:
        raise NotImplementedError

    def request_named_send_async(self, sender_id: Hashable, recipient_id: Hashable, message: Any) -> NoReturn:
        raise NotImplementedError

    def request_named_send_sync(self, sender_id: Hashable, recipient_id: Hashable, message: Any) -> NoReturn:
        raise NotImplementedError

    def request_named_read_async(self, recipient_id: Hashable) -> Tuple[Hashable, Any]:
        """
        Will return tuple with sender ID and message

        :rtype: Tuple[Hashable, Any]
        """
        raise NotImplementedError

    def request_named_read_sync(self, recipient_id: Hashable) -> NoReturn:
        raise NotImplementedError

    def full_processing_iteration(self):
        pass

    def in_work(self) -> bool:
        return self.thrifty_in_work(False)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['cli_selector', 'acli_selector']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.asyncio.ainput import ainput

from enum import IntEnum
from typing import Dict, Type, Optional


class CliSelectorTerminatedByUser(Exception):
    pass


def cli_selector(
        variants_type: Type[IntEnum], 
        variant_type_name_per_variant_type: Dict[IntEnum, str], 
        prompt: str,
        default_index: int = 1,
        header: Optional[str] = None, 
        result_template: Optional[str] = None, 
        ):
    if 1 > default_index:
        default_index = 1
    
    variants_num: int = len(variants_type)
    if 1 > variants_num:
        raise ValueError('Empty variants_type')
    
    if default_index > variants_num:
        default_index = variants_num
    
    default_index_str = str(default_index)

    if header is not None:
        print(header)

    for variant_type, variant_type_name in variant_type_name_per_variant_type.items():
        print(f'{variant_type.value}. {variant_type_name}')
    
    result = None
    while not isinstance(result, variants_type):
        try:
            result = variants_type(int((input(prompt)).strip() or default_index_str))
        except ValueError:
            pass  # Try again.
        except KeyboardInterrupt:
            continue  # Control-C pressed. Try again.
        except EOFError:
            raise CliSelectorTerminatedByUser() from None  # Control-D pressed.
    
    if result_template is not None:
        print(result_template.format(result))
    
    return result


async def acli_selector(
        variants_type: Type[IntEnum], 
        variant_type_name_per_variant_type: Dict[IntEnum, str], 
        prompt: str,
        default_index: int = 1,
        header: Optional[str] = None, 
        result_template: Optional[str] = None, 
        ):
    if 1 > default_index:
        default_index = 1
    
    variants_num: int = len(variants_type)
    if 1 > variants_num:
        raise ValueError('Empty variants_type')
    
    if default_index > variants_num:
        default_index = variants_num
    
    default_index_str = str(default_index)

    if header is not None:
        print(header)

    for variant_type, variant_type_name in variant_type_name_per_variant_type.items():
        print(f'{variant_type.value}. {variant_type_name}')
    
    result = None
    while not isinstance(result, variants_type):
        try:
            result = variants_type(int((await ainput(prompt)).strip() or default_index_str))
        except ValueError:
            pass  # Try again.
        except KeyboardInterrupt:
            continue  # Control-C pressed. Try again.
        except EOFError:
            raise CliSelectorTerminatedByUser() from None  # Control-D pressed.
    
    if result_template is not None:
        print(result_template.format(result))
    
    return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import getouterframes, currentframe
from cengal.web_tools.request_cache import RequestCache
from typing import Optional
from cengal.introspection.inspect import frame
import os

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class LineTracer:
    def __init__(self, print_full_file_name=False, trace_allowed=True):
        self.print_full_file_name = print_full_file_name
        self.trace_allowed = trace_allowed
        self._file_cache = RequestCache(999999)
        pass

    def _get_file_data(self, path):
        file_data = self._file_cache.try_to_get_data_for_request(path)
        if not file_data:
            with open(path) as file:
                file_data = file.readlines()
                file_data = [line.rstrip('\n') for line in file_data]
                self._file_cache.put_new_request(path, file_data)
        return file_data

    def _get_file_line(self, path, line_num):
        line_num -= 1  # make line numbers start from 0 - not from 1
        default_answer = None
        if line_num < 0:
            return default_answer
        
        file_data = self._get_file_data(path)
        max_line = len(file_data) - 1
        if line_num > max_line:
            return default_answer

        return file_data[line_num]
    
    def _frame_info(self, depth: Optional[int] = 1):
        fr = frame(depth + 1)
        filename = fr.f_code.co_filename
        line_number = fr.f_lineno
        function_name = fr.f_code.co_name
        lines = str()
        index = fr.f_code.co_firstlineno
        return fr, filename, line_number, function_name, lines, index

    def trace(self, depth: Optional[int] = 1):
        if not self.trace_allowed:
            result = (None, None, None, None, None)
            return result
        
        frame, filename, line_number, function_name, lines, index = self._frame_info(depth + 1)
        previous_line_num = line_number - 1
        lines = self._get_file_line(filename, previous_line_num)
        if not self.print_full_file_name:
            filename = os.path.basename(filename)
        
        result = (filename, function_name, previous_line_num, lines, index)
        return result

    def trace_self(self, depth: Optional[int] = 1):
        if not self.trace_allowed:
            result = (None, None, None, None, None)
            return result
        
        frame, filename, line_number, function_name, lines, index = self._frame_info(depth + 1)
        current_line_num = line_number
        lines = self._get_file_line(filename, current_line_num)
        if not self.print_full_file_name:
            filename = os.path.basename(filename)
        
        result = (filename, function_name, current_line_num, lines, index)
        return result

    def trace_next(self, depth: Optional[int] = 1):
        if not self.trace_allowed:
            result = (None, None, None, None, None)
            return result
        
        frame, filename, line_number, function_name, lines, index = self._frame_info(depth + 1)
        next_line_num = line_number + 1
        lines = self._get_file_line(filename, next_line_num)
        if not self.print_full_file_name:
            filename = os.path.basename(filename)
        
        result = (filename, function_name, next_line_num, lines, index)
        return result

    def trace_exact(self, line_num: int, depth: Optional[int] = 1):
        if not self.trace_allowed:
            result = (None, None, None, None, None)
            return result
        
        frame, filename, line_number, function_name, lines, index = self._frame_info(depth + 1)
        exact_line_num = line_num
        lines = self._get_file_line(filename, exact_line_num)
        if not self.print_full_file_name:
            filename = os.path.basename(filename)
        
        result = (filename, function_name, exact_line_num, lines, index)
        return result

    def trace_relative(self, line_num_offset: int, depth: Optional[int] = 1):
        if not self.trace_allowed:
            result = (None, None, None, None, None)
            return result
        
        frame, filename, line_number, function_name, lines, index = self._frame_info(depth + 1)
        exact_line_num = line_number + line_num_offset
        lines = self._get_file_line(filename, exact_line_num)
        if not self.print_full_file_name:
            filename = os.path.basename(filename)
        
        result = (filename, function_name, exact_line_num, lines, index)
        return result

    def __call__(self, depth: Optional[int] = 1):
        if not self.trace_allowed:
            return
        
        print(self.trace(depth=depth))

    def n(self, name: str = '', depth: Optional[int] = 1):
        if not self.trace_allowed:
            return
        
        if name:
            print(f'<< {name} >>', self.trace(depth=depth))
        else:
            print(self.trace(depth=depth))


line_tracer = LineTracer()
line_tracer_full_file_name = LineTracer(print_full_file_name=True)


def previous_line_number(depth: Optional[int] = 1) -> int:
    return line_tracer.trace(depth + 1)[2]


def current_line_number(depth: Optional[int] = 1) -> int:
    return previous_line_number(depth + 1) + 1


def next_line_number(depth: Optional[int] = 1) -> int:
    return previous_line_number(depth + 1) + 2


pln = previous_line_number
cln = current_line_number
nln = next_line_number
